from RPA.application import (
    BaseApplication,
    catch_com_error,
    constants,
    to_path,
    to_str_path,
)
from enum import Enum
from typing import Any


class CursorPosition(Enum):
    """Enum for moving cursor position"""

    NO_MOVE = 0
    START = 1
    END = 2


class Application(BaseApplication):
    """`Word.Application` is a library for controlling the Word application.

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library                 RPA.Word.Application
        Task Setup              Open Application
        Suite Teardown          Quit Application

        *** Tasks ***
        Open existing file
            Open File           old.docx
            Write Text          Extra Line Text
            Write Text          Another Extra Line of Text
            Save Document AS    ${CURDIR}${/}new.docx
            ${texts}=           Get all Texts
            Close Document

    **Python**

    .. code-block:: python

        from RPA.Word.Application import Application

        app = Application()
        app.open_application()
        app.open_file('old.docx')
        app.write_text('Extra Line Text')
        app.save_document_as('new.docx')
        app.quit_application()
    """

    APP_DISPATCH = "Word.Application"
    FILEFORMATS = {
        "DEFAULT": "wdFormatDocumentDefault",
        "PDF": "wdFormatPDF",
        "RTF": "wdFormatRTF",
        "HTML": "wdFormatHTML",
        "WORD97": "wdFormatDocument97",
        "OPENDOCUMENT": "wdFormatOpenDocumentText",
    }

    def open_file(self, filename: str, read_only: bool = True) -> None:
        """Open Word document with filename.

        :param filename: Word document path
        """
        path = to_path(filename)
        if not path.is_file():
            raise FileNotFoundError(f"{str(path)!r} doesn't exist")

        state = "read-only" if read_only else "read-write"
        self.logger.info("Opening document (%s): %s", state, path)
        with catch_com_error():
            doc = self.app.Documents.Open(
                FileName=str(path),
                ConfirmConversions=False,
                ReadOnly=read_only,
                AddToRecentFiles=False,
            )

        err_msg = None
        if doc is None:
            err_msg = (
                "Got null object when opening the document, enable RDP connection if"
                " running by Control Room through a Worker service"
            )
        elif not hasattr(doc, "Activate"):
            err_msg = (
                "The document can't be activated, open it manually and dismiss any"
                " alert you may encounter first"
            )
        if err_msg:
            raise IOError(err_msg)

        doc.Activate()
        self.app.ActiveWindow.View.ReadingLayout = False

    def create_new_document(self) -> None:
        """Create new document for Word application"""
        with catch_com_error():
            self.app.Documents.Add()

    def export_to_pdf(self, filename: str) -> None:
        """Export active document into PDF file.

        :param filename: PDF to export WORD into
        """
        path = to_str_path(filename)
        with catch_com_error():
            self._active_document.ExportAsFixedFormat(
                OutputFileName=path, ExportFormat=constants.wdExportFormatPDF
            )

    def replace_text(self, find: str, replace: str) -> None:
        """Replace text in active document

        :param find: text to replace
        :param replace: new text
        """
        self._active_document.Content.Find.Execute(FindText=find, ReplaceWith=replace)

    def set_header(self, text: str) -> None:
        """Set header for the active document

        :param text: header text to set
        """
        for section in self._active_document.Sections:
            for header in section.Headers:
                header.Range.Text = text

    def set_footer(self, text: str) -> None:
        """Set footer for the active document

        :param text: footer text to set
        """
        for section in self._active_document.Sections:
            for footer in section.Footers:
                footer.Range.Text = text

    def save_document(self) -> None:
        """Save active document"""
        # Accept all revisions
        self._active_document.Revisions.AcceptAll()
        # Delete all comments
        if self._active_document.Comments.Count >= 1:
            self._active_document.DeleteAllComments()

        self._active_document.Save()

    def save_document_as(self, filename: str, fileformat: str = None) -> None:
        """Save document with filename and optionally with given fileformat

        :param filename: where to save document
        :param fileformat: see @FILEFORMATS dictionary for possible format,
            defaults to None
        """
        path = to_str_path(filename)

        # Accept all revisions
        self._active_document.Revisions.AcceptAll()
        # Delete all comments
        if self._active_document.Comments.Count >= 1:
            self._active_document.DeleteAllComments()

        if fileformat and fileformat.upper() in self.FILEFORMATS:
            self.logger.debug("Saving with file format: %s", fileformat)
            format_name = self.FILEFORMATS[fileformat.upper()]
            format_type = getattr(constants, format_name)
        else:
            format_type = constants.wdFormatDocumentDefault

        try:
            self._active_document.SaveAs2(path, FileFormat=format_type)
        except AttributeError:
            self._active_document.SaveAs(path, FileFormat=format_type)

        self.logger.info("File saved to: %s", path)

    def get_all_texts(self) -> str:
        """Get all texts from active document

        :return: texts
        """
        return self._active_document.Content.Text

    def paste_from_clipboard(self) -> None:
        """Paste content from clipboard to the document's
        current cursor position."""
        self.app.Selection.Paste()

    def get_current_line(self) -> str:
        """Get the text of the current line in the document."""
        original_range = self.app.Selection.Range
        self.app.Selection.Expand(Unit=constants.wdLine)
        text = self.app.Selection.Text
        original_range.Select()
        return text

    def get_number_of_lines(self) -> int:
        """Get the number of lines in the document."""
        if self.app.Documents.Count >= 1:
            return self.app.Documents.Item(1).ComputeStatistics(
                constants.wdStatisticLines
            )
        else:
            raise AssertionError("No document is open")

    def move_to_top(self) -> None:
        """Move cursor to the top of the document."""
        self.app.Selection.HomeKey(Unit=constants.wdStory)

    def move_to_end(self) -> None:
        """Move cursor to the end of the document."""
        self.app.Selection.EndKey(Unit=constants.wdStory)

    def move_to_line_start(self) -> None:
        """Move cursor to start of the line on the current cursor position."""
        self.app.Selection.HomeKey(Unit=constants.wdLine)

    def move_to_line_end(self) -> None:
        """Move cursor to end of the line on the current cursor position."""
        self.app.Selection.EndKey(Unit=constants.wdLine)

    def move_vertically(
        self,
        lines: int = 0,
    ) -> Any:
        """Move cursor vertically from current cursor position.

        Remember that if cursor is already at the top the cursor can't
        move up and if cursor is already at the bottom the cursor can't
        move down.

        :param lines: lines to move
        """
        if lines > 0:
            self.app.Selection.MoveDown(Unit=constants.wdLine, Count=lines)
        elif lines < 0:
            self.app.Selection.MoveUp(Unit=constants.wdLine, Count=-lines)
        self.app.Selection.Select()

    def move_horizontally(
        self,
        characters: int = 0,
    ) -> Any:
        """Move cursor horizontally from current cursor position.

        Remember that if cursor is already at the start the cursor can't move
        left and if cursor is already at the end the cursor can't move right.

        :param characters: characters to move
        """
        if characters > 0:
            self.app.Selection.MoveRight(Unit=constants.wdCharacter, Count=characters)
        elif characters < 0:
            self.app.Selection.MoveLeft(Unit=constants.wdCharacter, Count=-characters)
        self.app.Selection.Select()

    def find_text(
        self,
        text: str,
        cursor_position: CursorPosition = CursorPosition.NO_MOVE,
        copy: bool = False,
    ) -> None:
        """Find text in the document.

        :param text: text to find
        :param cursor_position: where to move cursor after finding text
        :param copy: copy found text into clipboard
        :raises AssertionError: if text is not found
        """
        found = False
        content_range = None
        if self.app.Documents.Count >= 1:
            content_range = self.app.Documents.Item(1).Content
            content_range.Find.ClearFormatting()
            found = content_range.Find.Execute(FindText=text)
        else:
            raise AssertionError("No document is open")
        if not found:
            raise AssertionError(f"Text '{text}' not found in the document")
        if copy:
            content_range.Copy()
        if cursor_position != CursorPosition.NO_MOVE:
            direction = (
                constants.wdCollapseStart
                if cursor_position == CursorPosition.START
                else constants.wdCollapseEnd
            )
            content_range.Collapse(Direction=direction)

    def write_text(
        self,
        text: str,
        cursor_position: CursorPosition = CursorPosition.NO_MOVE,
        end_of_text: bool = True,
    ) -> None:
        """Writes given text at the end of the document

        :param text: string to write
        :param cursor_position: where to move cursor before writing
        :param end_of_text: if True moves cursor to the end of the text
         before writing
        """

        if end_of_text:
            self.move_to_end()

        if cursor_position != CursorPosition.NO_MOVE:
            direction = (
                constants.wdCollapseStart
                if cursor_position == CursorPosition.START
                else constants.wdCollapseEnd
            )

            self.app.Selection.Collapse(Direction=direction)

        self.app.Selection.TypeText(text)
        self.app.Selection.Select()

    def select_current_paragraph(self):
        """Select text in current active paragraph."""
        self.app.Selection.Paragraphs(1).Range.Select()

    def copy_selection_to_clipboard(self):
        """Copy current text selection to clipboard."""
        self.app.Selection.Copy()

    def select_paragraph(self, count: int = 1):
        """Select paragraph(s) from current cursor position.

        Negative `count` moves cursor up number of paragraphs and
        positive `count` moves cursor down number of paragraphs.

        :param count: number of paragraphs to select
        """
        start = self.app.Selection.Range.Start
        self.app.Selection.MoveDown(Unit=constants.wdParagraph, Count=count)
        end = self.app.Selection.Range.End
        self.app.Selection.SetRange(start, end)



import logging
import os
from collections import OrderedDict
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from robot.api.deco import keyword, library
from robot.utils.robottime import parse_time, timestr_to_secs

from smartsheet import Smartsheet as SmartSDK
from smartsheet.models import (
    Attachment,
    Cell,
    Column,
    IndexResult,
    Row,
    ServerInfo,
    Sheet,
    SheetFilter,
    User,
)

# isort: off
# isort is not capable of sorting the try block correctly
from RPA.Robocorp.utils import PathType, get_output_dir
from RPA.version import __version__

try:
    from RPA.Tables import Table

    TABLES = True
except ImportError:
    TABLES = False
# isort: on


CommaListType = Union[str, List[Any]]
ColumnType = Union[Dict, Column]
RowType = Union[OrderedDict, Row]
if TABLES:
    TableType = Table
else:
    TableType = Any


class SmartsheetError(Exception):
    """Base error class for Smartsheet library."""


class SmartsheetAuthenticationError(SmartsheetError):
    """Error when authenticated Smartsheet instance does not exist."""


class SmartsheetNoSheetSelectedError(SmartsheetError):
    """Error when no sheet was selected to on which to perform the operation."""


def _require_auth(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.smart is None:
            raise SmartsheetAuthenticationError(
                "Authentication was not completed. "
                "Please use the Set access token keyword to authenticate."
            )
        return func(self, *args, **kwargs)

    return wrapper


@library(scope="GLOBAL", doc_format="REST")
class Smartsheet:
    """*Smartsheet* is a library for accessing Smartsheet using the
    `Smartsheet API 2.0`_. It extends `smartsheet-python-sdk`_.

    .. _Smartsheet API 2.0: https://smartsheet.redoc.ly/
    .. _smartsheet-python-sdk: https://github.com/smartsheet/smartsheet-python-sdk

    Getting started
    ===============

    To use this library, you need to have a Smartsheet account and an API token.
    You can get your API token from the `Smartsheet Developer Portal`_.
    This library currently only supports raw token authentication. Once
    obtained, you can configure the access token using the ``Set Access Token``
    keyword or via the ``access_token`` argument in the library import.

    .. _Smartsheet Developer Portal: https://smartsheet-platform.github.io/api-docs/

    Working on a sheet
    ==================

    The library supports working on a single sheet at a time. To select a sheet
    to work on, use the ``Select Sheet`` keyword. This will set the sheet as
    the active sheet for all subsequent operations. Some operations
    update the sheet, but this will not necessarily be reflected in the active
    sheet. To refresh the active sheet, use the ``Refresh Sheet`` keyword.

    Native Smartsheet objects
    =========================

    You can retrieve the native Smartsheet object from many keywords by
    specifying the ``native`` argument. The default will return a more
    common Python object, such as a dictionary or list. The native object
    is a class from the `smartsheet-python-sdk`_ library and will have
    additional methods and attributes. The most important attributes
    available for most native objects are (some may be unavailable
    for some objects):

    - ``id``: the unique identifier of the object
    - ``name``: the name of the object
    - ``title``: the title of a column
    - ``permalink``: the URL to the object
    """

    # TODO: ADD LOGGING THROUGHOUT

    # Only implemented includes are available, others exist for the API
    SHEET_INCLUDES = {
        "rowId": None,  # function must be set in __init__
        "attachments": "attachments",
        "attachmentFiles": None,  # function must be set in __init__
        "discussions": "discussions",
        "rowPermalink": "permalink",
    }
    """Dictionary of supported additional objects to ask for via ``include``
    parameters for sheets. Key is the name of the option, value is the
    API attribute or function to call.
    """
    ROW_INCLUDES = {
        "rowId": None,  # function must be set in __init__
        "attachments": "attachments",
        "attachmentFiles": None,  # function must be set in __init__
        "discussions": "discussions",
        "rowPermalink": "permalink",
        "filters": "filteredOut",
    }
    """Dictionary of supported additional objects to ask for via ``include``
    parameters for rows. Key is the name of the option, value is the
    API attribute or function to call.
    """
    ROW_EXCLUDES = [
        "filteredOutRows",
        "linkInFromCellDetails",
        "linksOutToCellsDetails",
        "nonexistentCells",
    ]
    """List of supported options to remove cells from rows."""
    SEARCH_INCLUDES = {"favoriteFlag": None}
    """Dictionary of supported additional objects to ask for via ``include``
    parameters for searches. Key is the name of the option, value is the
    API attribute or function to call.
    """
    SEARCH_SCOPES = [
        "attachments",
        "cellData",
        "comments",
        "folderNames",
        "reportNames",
        "sheetNames",
        "sightNames",
        "summaryFields",
        "templateNames",
        "workspaceNames",
    ]

    def __init__(self, access_token: str = None, max_retry_time: Union[str, int] = 30):
        """If you do not initialize the library with an access token,
        it will attempt to load the environment variable
        ``SMARTSHEET_ACCESS_TOKEN``, otherwise, use the keyword
        ``Set access token``.

        :param access_token: The access token created for your
         Smartsheet user.
        :param max_retry_time: Maximum time to allow retries of API
         calls. Can be provided as a time string or int.
        """
        self.logger = logging.getLogger(__name__)
        # Sets log level for the SDK to debug, we will redirect their logs here
        # and let Robot Framework handle logging levels once here.
        os.environ["LOG_CFG"] = "DEBUG"
        self.smart = None
        self.max_retry_time = None
        self.set_max_retry_time(max_retry_time)
        self._set_token(access_token)
        # sheet init
        self._sheets = []
        self.current_sheet = None
        # sets the function for this key
        self.SHEET_INCLUDES["attachmentFiles"] = self._download_row_attachments
        self.SHEET_INCLUDES["rowId"] = self._parse_row_id
        self.ROW_INCLUDES["attachmentFiles"] = self._download_row_attachments
        self.ROW_INCLUDES["rowId"] = self._parse_row_id
        # user init
        self.current_user = None
        # Server init
        self.app_constants = None

    # *** AUTH AND CONFIG ***
    def _set_token(self, access_token: Optional[str] = None):
        # The env var is used by the underlying client library if it exists. The
        # underlying library will throw an error if the env var is missing and no
        # access token was passed in.
        if (
            os.environ.get("SMARTSHEET_ACCESS_TOKEN", None) is None
            and access_token is None
        ):
            self.smart = None
        else:
            self.smart = SmartSDK(
                access_token=access_token,
                user_agent=f"rpaframework/{__version__}",
                max_retry_time=self.max_retry_time,
            )
            self.smart.errors_as_exceptions(True)
            # TODO: consider if this properly connects the logs.
            for handler in logging.getLogger("smartsheet").handlers:
                self.logger.addHandler(handler)

    def _require_current_sheet(self) -> None:
        if self.current_sheet is None:
            raise SmartsheetNoSheetSelectedError(
                "You must select a sheet before performing this operation."
            )

    @keyword
    def set_access_token(self, access_token: str) -> None:
        """Sets the access token to be used when accessing the
        Smartsheet API.

        Learn more about authenticating to Smartsheets
        `here <https://smartsheet.redoc.ly/#section/API-Basics/Raw-Token-Requests>`_.

        :param access_token: The access token created for your
         Smartsheet user.

        Example:

        .. code-block:: robotframework

            Set Access Token  ${access_token}

        .. code-block:: python

            smartsheet = Smartsheet(access_token=access_token)
            # or
            smartsheet.set_access_token(access_token)
        """
        self._set_token(access_token)

    @keyword
    def set_max_retry_time(self, max_retry_time: Union[str, int]) -> Optional[int]:
        """Sets the max retry time to use when sending requests to the
        Smartsheet API. Returns the current max retry time.

        :param max_retry_time: Maximum time to allow retries of API
         calls. Can be provided as a time string or int.
        """
        old_time = self.max_retry_time
        self.max_retry_time = int(timestr_to_secs(max_retry_time, round_to=None))
        if self.smart is not None:
            # pylint: disable=protected-access
            # might be affected by smart's def of __getattr__ ... this is a patch
            # to allow modification of the calculated backoff time after init.
            self.smart._user_calc_backoff._max_retry_time = self.max_retry_time
        return old_time

    # *** UTILITY ***
    def _parse_comma_list_type(
        self, comma_list: Optional[CommaListType] = None
    ) -> Optional[List[str]]:
        if comma_list is None:
            return None
        elif isinstance(comma_list, str):
            return [s.strip() for s in comma_list.split(",")]
        elif isinstance(comma_list, list):
            return [str(s) for s in comma_list]
        else:
            raise TypeError(
                "Invalid type supplied as comma list, either provide a "
                "list object or a string of comma-separated values."
            )

    def _unpack_index_result(
        self, index_result: Optional[IndexResult] = None
    ) -> Optional[List[object]]:
        """Unpacks the provided IndexResult object into a list of the
        underlying type.
        """
        return index_result.data if index_result else None

    def _parse_include(
        self, include_type: Dict, include: Optional[CommaListType] = None
    ) -> Optional[List[str]]:
        """Parses the include parameter based on the provided include type."""
        include = self._parse_comma_list_type(include)
        if include is not None:
            includes = []
            for i in include:
                if str(i).upper() == "ALL":
                    includes = [str(k) for k in include_type]
                    break
                elif i in include_type:
                    includes.append(str(i))
                else:
                    raise ValueError(
                        f"Invalid or not implemented value(s) for include: {i}"
                    )
        else:
            includes = None
        return includes

    def _parse_exclude(
        self, exclude_type: List, exclude: CommaListType = None
    ) -> Optional[List[str]]:
        """Parses the exclude parameter based on the provided exclude type."""
        exclude = self._parse_comma_list_type(exclude)
        if exclude is not None and any(i not in exclude_type for i in exclude):
            raise ValueError("Invalid or not implemented value(s) for exclude.")
        return exclude

    def _parse_scope(
        self, scope_type: List, scope: Optional[CommaListType] = None
    ) -> List[str]:
        """Parses the scope parameter based on the provided scope type."""
        scope = self._parse_comma_list_type(scope)
        if scope is not None and any(i not in scope_type for i in scope):
            raise ValueError("Invalid or not implemented value(s) for scope.")
        return scope

    @keyword
    @_require_auth
    def get_application_constants(self) -> ServerInfo:
        """Gets application constants from the server. This is not
        necessary for most automation scenarios, but may be useful for
        debugging or for other advanced scenarios.
        """
        self.app_constants = self.smart.Server.server_info()
        return self.app_constants

    # *** SHEETS ***
    @property
    def sheets(self) -> List[Sheet]:
        """Full list of cached sheets."""
        if len(self._sheets) == 0:
            self._refresh_sheets()
        return self._sheets

    @_require_auth
    def _refresh_sheets(self):
        """Refreshes the ``sheets`` property."""
        self._sheets = self.smart.Sheets.list_sheets(include_all=True).data

    def _find_sheet(self, name: str) -> Optional[IndexResult]:
        """Finds a sheet by name and returns the ``IndexResult``, which
        can be used to pull sheet data with its ``id`` attribute.
        """
        for sheet in self.sheets:
            if sheet.name == name:
                return sheet
        return None

    def _parse_sheet_id(
        self, sheet_id: Optional[int] = None, sheet_name: Optional[str] = None
    ) -> Optional[int]:
        """Gets the sheet ID between the provided ID and/or Name."""
        if sheet_id is not None and sheet_name is not None:
            raise ValueError("You cannot provide both sheet_id and sheet_name.")
        if sheet_id is None and sheet_name is None:
            raise ValueError("You must provide either sheet_id or sheet_name.")
        if sheet_id is not None:
            return sheet_id
        if sheet_name is not None:
            try:
                return self._find_sheet(sheet_name).id
            except AttributeError as e:
                raise ValueError(f"Sheet '{sheet_name}' not found.") from e
        return None

    @keyword
    @_require_auth
    def list_sheets(self, use_cache: bool = True) -> List[Sheet]:
        """Lists all sheets available for the authenticated account. Uses
        cached lists if available unless ``use_cache`` is set to ``False``.

        The cached lists is used for other keywords, so if you need to
        refresh the cache for other keywords to use, you must do so
        via this keyword.

        :param use_cache: Defaults to ``True``. You can set to ``False``
         to force a reload of the cached list of sheets.

        Example:

        .. code-block:: robotframework

            ${sheets}=  List Sheets
            FOR  ${sheet}  IN  @{sheets}
                Log  ${sheet.name}

        .. code-block:: python

            ss = SmartsheetLibrary(account_token=account_token)
            sheets = ss.list_sheets()
            for sheet in sheets:
                print(sheet.name)
        """
        if not use_cache:
            self._refresh_sheets()

        return self.sheets

    @keyword
    def unselect_current_sheet(self) -> None:
        """Resets the current sheet to `None`."""
        self.current_sheet = None

    @keyword
    @_require_auth
    def get_sheet(
        self,
        sheet_id: Optional[int] = None,
        sheet_name: Optional[str] = None,
        include: Optional[CommaListType] = None,
        row_ids: Optional[CommaListType] = None,
        row_numbers: Optional[CommaListType] = None,
        column_ids: Optional[CommaListType] = None,
        filter_id: Optional[int] = None,
        native: bool = False,
        download_path: Optional[PathType] = None,
    ) -> Union[TableType, Sheet]:
        """Retrieves a sheet from Smartsheet. This keyword also sets
        the currently selected sheet to the returned sheet.

        You cannot provide both a ``sheet_id`` and ``sheet_name``.

        :param sheet_id: The ID of the sheet to get. You cannot supply
         both a ``sheet_id`` and ``sheet_name``.
        :param sheet_name: The name of the sheet to get, it will return
         the first sheet name matching the provided name. You cannot supply
         both a ``sheet_id`` and ``sheet_name``.
        :param include: Additional metadata which can be retrieved with
         the table. The list can only contain the following items:
         ``attachments``, ``attachmentFiles``, ``discussions``,
         ``rowPermalink``, or ``ALL``. Note that ``attachmentFiles``
         will only download files if you do not set ``native`` to
         ``True``.
        :param row_ids: A list of row IDs to include. All other rows
         will be ignored. The list can be a list object or a
         comma-separated list as a string.
        :param row_numbers: A list of row numbers to include. All other
         rows will be ignored. The list can be a list object or a
         comma-separated list as a string.
        :param column_ids: A list of column IDs to only include, others
         will not be returned. The list can be a list object or a
         comma-separated list as a string.
        :param filter_id: The ID of a filter to apply. Filtered out
         rows will not be included in the resulting table.
        :param native: Defaults to ``False``. Set to ``True`` to change
         the return type to the native Smartsheet data model. The native
         type is useful for passing to other keywords as arguments.
        :param download_path: Defaults to ``None``. Can be set when
         ``attachmentFiles`` is included in the ``include`` parameter.
         All attachments will be downloaded to the provided directory.

        Example:

        .. code-block:: robotframework

            ${sheet}=  Get Sheet  sheet_name=My Sheet
            FOR  ${row}  IN  &{sheet}
                FOR  ${column}  ${value}  IN  &{row}
                    Log  The column ${column} has the value ${value}
                END
            END

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            sheet = ss.get_sheet(sheet_name="My Sheet", native=True)
            for row in sheet:
                for cell in row:
                    print(f"The column {cell.column_id} has the value {cell.value}")
        """
        sheet_id = self._parse_sheet_id(sheet_id, sheet_name)
        # Smartsheet class defines a `__getattr__` which makes the below work, just
        # doesn't help with intellisense autocompletion.
        includes = self._parse_include(self.SHEET_INCLUDES, include)
        sheet = self.smart.Sheets.get_sheet(
            sheet_id,
            include=includes,
            row_ids=self._parse_comma_list_type(row_ids),
            row_numbers=self._parse_comma_list_type(row_numbers),
            exclude="filteredOutRows",
            column_ids=self._parse_comma_list_type(column_ids),
            filter_id=filter_id,
        )
        sheet.selected_includes = includes
        if download_path is not None:
            sheet.download_path = download_path
        self.current_sheet = sheet
        if native:
            return sheet
        else:
            return self.convert_sheet_to_table(sheet)

    @keyword
    def convert_sheet_to_table(self, sheet: Optional[Sheet] = None) -> Table:
        """Converts the current sheet to table. You can provide a differnt
        native sheet object to be converted via the ``sheet`` parameter.

        This keyword attempts to return the sheet as a table via
        ``RPA.Tables``, but if that library is not available in this
        context, the sheet is returned as its native data model (e.g.,
        no operation is performed).

        If the sheet contains additional data from the ``include``
        argument, they will be appended to the end of the table as
        additional columns in the data model. These additional objects
        will be attached as a list of objects depending on the items
        requested.

        .. warn:

            Only the following supported additional data requests
            are supported when unpacking a Sheet object.

            * ``attachments``: The value for the cell will be a list of
            native Attachment object.
            * ``attachmentFiles``: The value for the cell will be a list of
            file paths to the local copies of each attachment. This
            operation may be resource intensive for many attachments.
            * ``discussions``: The value for the cell will be a list of
            native Discussion objects.
            * ``rowPermalink``: The value for the cell will be a single
            string representing the URL to the permalink.
        """
        if sheet is None:
            self._require_current_sheet()
            sheet = self.current_sheet
        includes = getattr(sheet, "selected_includes", []) or []
        table_data = []
        headers = []
        for column in sheet.columns:
            headers.append(column.title)
        if len(includes) > 0:
            headers.extend(includes)

        for row in sheet.rows:
            row_data = []
            for cell in row.cells:
                row_data.append(cell.value)
            for key in includes:
                key_or_func = self.SHEET_INCLUDES[key]
                if callable(key_or_func):
                    row_data.append(key_or_func(row))
                elif hasattr(row, key_or_func):
                    row_data.append(getattr(row, key_or_func, None))

            table_data.append(row_data)

        table = Table(table_data, columns=headers)

        return table

    @keyword
    def refresh_sheet(self, native: bool = False) -> Union[TableType, Sheet]:
        """Refreshes the current sheet from the API and returns it
        either as a Table or native data model depending on the
        ``native`` argument.
        """
        self._require_current_sheet()
        sheet = self.smart.Sheets.get_sheet(self.current_sheet.id)
        self.current_sheet = sheet
        if native:
            return sheet
        else:
            return self.convert_sheet_to_table(sheet)

    @keyword
    def get_sheet_owner(
        self, sheet_id: Optional[int] = None, sheet_name: Optional[str] = None
    ) -> Tuple[str, int]:
        """Returns the owner's username and ID for the current sheet."""
        sheet_id = self._parse_sheet_id(sheet_id, sheet_name)
        sheet = self.smart.Sheets.get_sheet(sheet_id, include=["ownerInfo"])
        return sheet.owner, sheet.owner_id

    @keyword
    def list_sheet_filters(
        self, sheet_id: Optional[int] = None, sheet_name: Optional[str] = None
    ) -> Optional[List[SheetFilter]]:
        """Returns a list of available filters for the current sheet. You
        can specify a different sheet via the ``sheet_id`` or
        ``sheet_name`` parameters.

        The returned list of filters can be used with the ``filter_id``
        argument of the ``get_sheet`` keyword.

        Example:

        .. code-block:: robotframework

            ${filters}=  List Sheet Filters
            FOR  ${filter}  IN  @{filters}
                ${filtered_sheet}=  Get Sheet
                ...  sheet_name=My sheet
                ...  filter_id=${filter.id}
                Log  There are ${len(filtered_sheet)} rows in the filtered sheet
            END

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            my_sheet_id = 123456789
            filters = ss.list_sheet_filters()
            for filter in filters:
                filtered_sheet = ss.get_sheet(
                    sheet_id=my_sheet_id,
                    filter_id=filter.id,
                    native=True,
                )
                print(
                    f"There are {len(filtered_sheet.rows)} rows in the "
                    f"filtered sheet"
                )
        """
        self._require_current_sheet()
        sheet_id = self._parse_sheet_id(sheet_id, sheet_name)
        filter_result = self.smart.Sheets.list_filters(
            sheet_id or self.current_sheet.id, include_all=True
        )
        return self._unpack_index_result(filter_result)

    @keyword
    @_require_auth
    def create_sheet(
        self,
        name: str,
        columns: Optional[List[ColumnType]] = None,
        from_sheet_id: Optional[Union[int, str]] = None,
    ) -> Sheet:
        """Creates a new sheet with the given name and columns, then sets
        the current sheet to the new sheet and returns it as a native
        Smartsheet object.

        :param name: Name of the new sheet.
        :param columns: List of columns to create in the new sheet.
        :param from_sheet_id: Sheet ID to use as a template for the new
         sheet.

        Example:

        .. code-block:: robotframework

            ${columns}=  Create List  Name  Email
            ${sheet}=  Create Sheet  My new sheet  ${columns}

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            columns = [
                {"title": "Name", "type": "TEXT_NUMBER"},
                {"title": "Email", "type": "TEXT_NUMBER"},
            ]
            sheet = ss.create_sheet("My new sheet", columns)
        """
        if columns is not None:
            new_columns = self._parse_columns(columns)
            sheet = self.smart.Home.create_sheet(
                Sheet({"name": name, "columns": new_columns})
            )
        elif from_sheet_id is not None:
            sheet = self.smart.Home.get_sheet({"name": name, "from_id": from_sheet_id})
        else:
            raise ValueError("Either columns or from_sheet_id must be provided.")
        self.current_sheet = sheet.result
        return sheet.result

    @keyword
    @_require_auth
    def search(
        self,
        query: str,
        location: Optional[str] = None,
        modified_since: Optional[Union[str, int]] = None,
        include: Optional[CommaListType] = None,
        scopes: Optional[CommaListType] = None,
    ) -> List[Sheet]:
        """Searches for all sheets for text matching the query. Returns
        a list of native Smartsheet Sheet objects. You can use the
        additional parameters to filter the search and increase speed.

        :param query: The text to search for.
        :param location: The location to search. When specified with
         a value of ``personalWorkspace``, the search will be limited
         to the current user's personal workspace.
        :param modified_since: The date to search from. This can be
         either a string or an integer. If an integer is provided, it
         will be interpreted as a Unix timestamp. If a string is
         provided, it will be parsed via the Robot Framework time
         utilities, so you can provided it using keywords like
         ``NOW - 1 day``.
        :param include: When specified with the value of ``favoriteFlag``,
         results will either include a ``favorite`` attribute or
         ``parentObjectFavorite`` attribute depending on the type of
         object found by the search engine.
        :param scopes: If search fails, try using an array for each type
         of this comma-separated list of search filters. The following
         strings can be used to filter the search results:

            * ``attachments``: Search in attachments.
            * ``cellData``: Search in cell data.
            * ``comments``: Search in comments.
            * ``folderNames``: Search in folder names.
            * ``reportNames``: Search in report names.
            * ``sheetNames``: Search in sheet names.
            * ``sightNames``: Search in sight names.
            * ``summaryFields``: Search in summary fields.
            * ``templateNames``: Search in template names.
            * ``workspaceNames``: Search in workspace names.


        Example:

        .. code-block:: robotframework

            ${sheets}=  Search  my search query
            FOR  ${sheet}  IN  @{sheets}
                Log  ${sheet.name}

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            sheets = ss.search("my search query")
            for sheet in sheets:
                print(sheet.name)
        """
        include = self._parse_include(self.SEARCH_INCLUDES, include)
        scopes = self._parse_scope(self.SEARCH_SCOPES, scopes)
        if isinstance(modified_since, str):
            modified_since = parse_time(modified_since)
        if modified_since is not None:
            modified_since = datetime.fromtimestamp(modified_since)
        return self.smart.Search.search(
            query, include, location, modified_since, scopes
        ).results

    # *** COLUMNS ***
    def _get_column_id(self, column: Union[int, str, Column]) -> int:
        """Returns the column ID for a column name. Can accept a
        column ID and will simply return it.
        """
        self._require_current_sheet()

        if isinstance(column, int):
            return column
        elif isinstance(column, str):
            for sheet_col in self.current_sheet.columns:
                if sheet_col.title == column:
                    return sheet_col.id
        elif isinstance(column, Column):
            return column.id
        else:
            raise TypeError(
                f"Invalid column type. Received {type(column)}, expected int or str."
            )
        raise ValueError(f"Column '{column!r}' not found.")

    def _parse_column(self, column: ColumnType) -> Column:
        """Parses a column and returns a Column object."""
        if isinstance(column, Column):
            return column
        elif isinstance(column, dict):
            new_column = Column()
            for key, value in column.items():
                try:
                    if key == "column_type":
                        key = "type"
                    setattr(new_column, key, value)
                except AttributeError:
                    self.logger.warning(
                        f"Unable to set attribute '{key}' for column: {column}"
                    )
            return new_column
        else:
            raise TypeError(
                f"Invalid column type. Received {type(column)}, "
                f"expected dict or Column."
            )

    def _parse_columns(self, columns: List[ColumnType]) -> List[Column]:
        """Parses a list of columns and returns a list of Column objects."""
        new_columns = []
        for column in columns:
            new_columns.append(self._parse_column(column))
        return new_columns

    @keyword
    def list_columns(
        self, sheet_id: Optional[int] = None, sheet_name: Optional[str] = None
    ) -> Optional[List[Column]]:
        """Returns a list of columns for the current sheet.

        :param sheet_id: The ID of the sheet to get columns from.
        :param sheet_name: The name of the sheet to get columns from.
        """
        sheet_id = self._parse_sheet_id(sheet_id, sheet_name)
        column_result = self.smart.Sheets.get_columns(sheet_id, include_all=True)
        return self._unpack_index_result(column_result)

    @keyword
    def add_columns(
        self,
        columns: List[ColumnType] = None,
    ) -> List[Column]:
        """Adds columns to the current sheet. Columns must be defined as
        a list of dictionaries or Column objects. Dictionaries can have
        additional keys set, see ``Add Column`` keyword for more information.

        Column types must be supported by the `Smartsheet API`_

        .. _Smartsheet API: https://smartsheet.redoc.ly/tag/columnsRelated#section/Column-Types

        :param columns: Columns as a list of dictionaries or Column
         objects.
        """  # noqa: E501
        self._require_current_sheet()

        new_columns = self._parse_columns(columns)
        column_response = self.smart.Sheets.add_columns(
            self.current_sheet.id, new_columns
        )
        return column_response.result

    @keyword
    def add_column(
        self,
        title: str,
        column_type: str = "TEXT_NUMBER",
        formula: Optional[str] = None,
        hidden: bool = False,
        index: Optional[int] = None,
        description: Optional[str] = None,
        primary: bool = False,
        format_string: Optional[str] = None,
        locked: bool = False,
        options: Optional[List[str]] = None,
        symbol: Optional[str] = None,
        validation: bool = False,
        width: Optional[int] = None,
    ):
        """Adds a column to the current sheet.

        :param title: Column title.
        :param column_type: Column type, must be a `supported type`_.
         Defaults to ``TEXT_NUMBER``.
        :param formula: Formula for the column (e.g., ``=data@row``).
         Defaults to ``None``.
        :param hidden: Whether the column is hidden. Defaults to ``False``.
        :param index: Index of the column. You can insert a column into
         and existing sheet by setting this index. Index is zero-based.
         Defaults to ``None`` which will add the column to the end of
         the sheet.
        :param description: Column description. Defaults to ``None``.
        :param primary: Whether the column is considered the primary
         key for indexing and searching. Defaults to ``False``.
        :param format_string: Column format using a `format descriptor`_
         string. Defaults to ``None``.
        :param locked: Whether the column is locked. Defaults to ``False``.
        :param options: List of options for a ``PICKLIST`` or
         ``MULTI_PICKLIST`` column. Defaults to ``None``.
        :param symbol: When a ``CHECKBOX`` or ``PICKLIST`` column has a
         display symbol, you can set the type of symbols by selected an
         appropriate string from the `symbol columns`_ definitions.
         Defaults to ``None``.
        :param validation: Whether validation has been enabled. Defaults
         to ``False``.
        :param width: Column width in pixels. Defaults to ``None``.

        .. _symbol columns: https://smartsheet.redoc.ly/tag/columnsRelated/#section/Column-Types/Symbol-Columns
        .. _format descriptor: https://smartsheet.redoc.ly/#section/API-Basics/Formatting
        .. _supported type: https://smartsheet.redoc.ly/tag/columnsRelated/#section/Column-Types

        Example:

        .. code-block:: robotframework

            Add Column  Title  TEXT_NUMBER
            Add Column  Description  TEXT_NUMBER  description=This is a description
            Add Column  Formula  TEXT_NUMBER  formula==data@row

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            ss.add_column(title="Title", column_type="TEXT_NUMBER")
            ss.add_column(title="Description", column_type="TEXT_NUMBER", description="This is a description")
            ss.add_column(title="Formula", column_type="TEXT_NUMBER", formula="=data@row")
        """  # noqa: E501
        self._require_current_sheet()
        return self.add_columns(
            [
                {
                    "title": title,
                    "type": column_type,
                    "formula": formula,
                    "hidden": hidden,
                    "index": index,
                    "description": description,
                    "primary": primary,
                    "format": format_string,
                    "locked": locked,
                    "options": options,
                    "symbol": symbol,
                    "validation": validation,
                    "width": width,
                }
            ]
        )

    @keyword
    def update_column(self, column: Union[int, str, Column], **kwargs) -> Column:
        """Updates a column in the current sheet. See the ``Add Column`` keyword
        for a list of supported attributes.

        :param column: Column ID or title.
        :param kwargs: Column attributes to update. See ``Add Column`` keyword
         for a list of supported attributes.
        """
        self._require_current_sheet()
        column_id = self._get_column_id(column)
        column = self.smart.Sheets.get_column(self.current_sheet.id, column_id)
        for key, value in kwargs.items():
            try:
                setattr(column, key, value)
            except AttributeError:
                self.logger.warning(f"Unable to set attribute '{key}' for column.")
        column_response = self.smart.Sheets.update_column(column)
        return column_response.result

    # *** ROWS ***
    def _parse_row_id(self, row: Union[int, Row]) -> int:
        """Returns the row ID from a row object or an integer. This
        function will search the current sheet for the row ID if an
        integer is provided, if it is not found, it will treat the
        row as a row number and search the sheet for the row ID. If
        not found there, it will return the row as an ID, assuming
        it is a valid but unknown row.
        """
        if isinstance(row, int):
            if row not in [r.id for r in self.current_sheet.rows]:
                return self._get_row_from_number(row)
            else:
                return row
        elif isinstance(row, Row):
            return row.id
        else:
            raise ValueError("Invalid row type.")

    def _create_row_from_dict(self, row_dict: Dict) -> Row:
        """Creates a row object from a dictionary."""
        row = Row()
        for key, value in row_dict.items():
            cell = Cell()
            if key in ["rowId", "row_id", "id"]:
                row.id = value
            elif key in ["rowNumber", "row_number"]:
                row.id = self._get_row_from_number(value)
            else:
                cell.column_id = self._get_column_id(key)
                cell.value = value
                row.cells.append(cell)
        return row

    def _get_row_from_number(self, row_number: int) -> int:
        """Returns the row ID from a row number. If not found, it
        assumes the provided row_number may have been an ID and
        returns it.
        """
        self._require_current_sheet()
        for row in self.current_sheet.rows:
            if row.row_number == row_number:
                return row.id
        return row_number

    def _create_row_from_list(self, row_list: List[Dict]) -> Row:
        """Creates a row object from a list."""
        row = Row()
        for cell_dict in row_list:
            if cell_dict.get("title") == "rowId":
                row.id = cell_dict.get("value")
            elif cell_dict.get("title") == "rowNumber":
                row.id = self._get_row_from_number(cell_dict.get("value"))
            else:
                cell = self._convert_dict_to_cell(cell_dict)
                row.cells.append(cell)
        return row

    def _convert_data_to_row(self, data: Union[Dict, List, Row]) -> Row:
        """Converts a dictionary or list to a row object."""
        if isinstance(data, dict):
            return self._create_row_from_dict(data)
        elif isinstance(data, list):
            return self._create_row_from_list(data)
        elif isinstance(data, Row):
            return data
        else:
            raise TypeError(
                f"Invalid row type. Received {type(data)}, expected dict, list or Row."
            )

    @keyword
    def convert_row_to_dict(self, row: Row) -> OrderedDict:
        """Converts a row object to a dictionary."""
        row_dict = OrderedDict()
        for cell in row.cells:
            row_dict[cell.column_id] = cell.value
        return row_dict

    @keyword
    def get_row(
        self,
        row: Union[int, Row],
        include: CommaListType = None,
        exclude: CommaListType = None,
        native: bool = False,
    ) -> Union[Row, OrderedDict]:
        """Returns a single row from the current sheet.

        You can provide the row as a native ``Row`` object or as an
        integer representing the row ID.
        """
        self._require_current_sheet()
        include = self._parse_include(self.ROW_INCLUDES, include)
        exclude = self._parse_exclude(self.ROW_EXCLUDES, exclude)
        row_id = self._parse_row_id(row)
        row = self.smart.Sheets.get_row(
            self.current_sheet.id, row_id, include=include, exclude=exclude
        )
        if native:
            return row
        else:
            return self.convert_row_to_dict(row)

    @keyword
    def set_rows(
        self,
        data: Union[List, Table],
        native: bool = False,
    ) -> List[RowType]:
        """Updates rows of the current sheet with the provided data.

        .. note::
           In order to update rows, you must identify the rows to the
           API. You can do this by providing the ``rowId`` or ``rowNumber``
           as a column in the data. The ID must be the API ID, while the
           number is the row number per the UI. This can only be excluded
           if the length of the data matches the length of the sheet.

        You can provide the data in several ways:

        * As a list of dictionaries: each list item represents a row
          as a single dictionary. The keys of the dictionary are the
          column IDs or Titles and the values are the values for the
          cells.
        * As a list of lists of dictionaries: each sub list item is
          a row and each dictionary is a cell. The keys of the dictionary
          should match cell attributes, e.g., ``column_id``, ``title``,
          ``value``, etc. See the `smartsheet API docs`_ for more
          information. The dictionary keys must be provided in snake
          case. You must use this method to set formulas on the row.
        * As a list of native ``Row`` objects: each ``Row`` object is a
          native object from the API with new values for the cells.
        * As a ``Table`` object: the columns of the Table must either
          be the column IDs or Titles.

        .. _smartsheet API docs: https://smartsheet.redoc.ly/tag/rows#operation/update-rows

        Examples:

        *Robot Framework:*

        .. code-block:: robotframework

            ${row1}=  Create Dictionary  rowId=123  column1=value1  column2=value2
            ${row2}=  Create Dictionary  rowId=456  column1=value3  column2=value4
            ${row3}=  Create Dictionary  rowId=789  column1=value5  column2=value6
            ${data} =  Create List  ${row1}  ${row2}  ${row3}
            Set Rows  ${data}

            # Or work with native row objects to update them.
            ${row1}=  Get Row  123
            FOR  ${cell}  IN  @{row1.cells}
                IF  ${cell.column_id} == 123
                    ${cell.value}=  Set Variable  New Value
                END
            END
            ${data}=  Create List  ${row1}
            Set Rows  ${data}

        *Python:*

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            row1 = {"rowId": 123, "column1": "value1", "column2": "value2"}
            row2 = {"rowId": 456, "column1": "value3", "column2": "value4"}
            row3 = {"rowId": 789, "column1": "value5", "column2": "value6"}
            data = [row1, row2, row3]
            ss.set_rows(data)

            # or work with native row objects to update them.
            row1 = ss.get_row(123)
            for cell in row1.cells:
                if cell.column_id == 123:
                    cell.value = "New Value"
            data = [row1]
            ss.set_rows(data)
        """  # noqa: E501
        self._require_current_sheet()
        new_rows = []
        # The Table will iterate as dictionaries
        for row in data:
            new_rows.append(self._convert_data_to_row(row))
        if len(new_rows) == len(self.current_sheet.rows):
            for i, row in enumerate(self.current_sheet.rows):
                new_rows[i].id = row.id
        row_update_response = self.smart.Sheets.update_rows(
            self.current_sheet.id, new_rows
        )
        self.logger.info(f"Updated {len(row_update_response.result)} rows.")
        if native:
            return row_update_response.result
        else:
            return [self.convert_row_to_dict(row) for row in row_update_response.result]

    @keyword
    def set_row(
        self,
        row: Union[int, Row],
        data: Union[Dict, List[Dict], Row] = None,
        native: bool = False,
    ) -> Union[Dict, Row]:
        """Updates a single row of the current sheet with the provided data.

        You can provide the row as a native ``Row`` object or as an
        integer representing the row ID. You may omit the ``data``
        argument if you are providing a native ``Row`` object.

        You can provide the data in several ways:

        * As a dictionary: the keys of the dictionary are the
          column IDs or Titles and the values are the values for the
          cells.
        * As a list of dictionaries: each dictionary is a cell. The keys
          of the dictionary should match cell attributes, e.g., ``column_id``,
          ``title``, ``value``, etc. See the `smartsheet API docs`_ for more
          information. The dictionary keys must be provided in snake case.
          You must use this method to set formulas on the row.
        * As a native ``Row`` object: a native object from the API with
          new values for the cells.

        .. _smartsheet API docs: https://smartsheet.redoc.ly/tag/rows#operation/update-rows

        For examples, see ``Set Rows``.
        """  # noqa: E501
        self._require_current_sheet()
        if isinstance(row, Row) and data is None:
            data = row
        row_id = self._parse_row_id(row)
        new_row = self._convert_data_to_row(data)
        new_row.id = row_id
        return self.set_rows([new_row], native=native)[0]

    @keyword
    def add_rows(self, data: Union[List, Table], native: bool = False) -> List[RowType]:
        """Adds rows to the current sheet with the provided data.

        You can provide the data in several ways:

        * As a list of dictionaries: each list item represents a row
          as a single dictionary. The keys of the dictionary are the
          column IDs or Titles and the values are the values for the
          cells.
        * As a list of lists of dictionaries: each sub list item is
          a row and each dictionary is a cell. The keys of the dictionary
          should match cell attributes, e.g., ``column_id``, ``title``,
          ``value``, etc. See the `smartsheet API docs`_ for more
          information. The dictionary keys must be provided in snake
          case. You must use this method to set formulas on the row.
        * As a list of native ``Row`` objects: each ``Row`` object is a
          native object from the API with new values for the cells.
        * As a ``Table`` object: the columns of the Table must either
          be the column IDs or Titles.

        .. _smartsheet API docs: https://smartsheet.redoc.ly/tag/rows#operation/add-rows

        Examples:

        *Robot Framework:*

        .. code-block:: robotframework

            ${row1}=  Create Dictionary  column1=value1  column2=value2
            ${row2}=  Create Dictionary  column1=value3  column2=value4
            ${row3}=  Create Dictionary  column1=value5  column2=value6
            ${data} =  Create List  ${row1}  ${row2}  ${row3}
            Add Rows  ${data}

        *Python:*

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            row1 = {"column1": "value1", "column2": "value2"}
            row2 = {"column1": "value3", "column2": "value4"}
            row3 = {"column1": "value5", "column2": "value6"}
            data = [row1, row2, row3]
            ss.set_rows(data)
        """
        self._require_current_sheet()
        new_rows = []
        # Table will iterate as dictionaries
        for row in data:
            new_rows.append(self._convert_data_to_row(row))
        row_update_response = self.smart.Sheets.add_rows(
            self.current_sheet.id, new_rows
        )
        self.logger.info(f"Added {len(row_update_response.result)} rows.")
        if native:
            return row_update_response.result
        else:
            return [self.convert_row_to_dict(row) for row in row_update_response.result]

    # *** CELLS ***
    def _convert_dict_to_cell(self, cell: Dict) -> Cell:
        """Converts a dictionary to a Cell object."""
        new_cell = Cell()
        if isinstance(cell, dict):
            for key, value in cell.items():
                if key == "title":
                    new_cell.column_id = self._get_column_id(value)
                else:
                    try:
                        setattr(new_cell, key, value)
                    except AttributeError as e:
                        raise KeyError(
                            f"Invalid cell dictionary key, '{key}' does not match any "
                            f"cell attributes."
                        ) from e
        else:
            raise TypeError(f"Invalid cell type. Received {type(cell)}, expected dict.")
        return new_cell

    @keyword
    def get_cell_history(
        self, row: Union[int, Row], column: Union[int, str, Column]
    ) -> Optional[List[Cell]]:
        """Retrieves the history of a cell in a row of the current sheet.

        :param row: The row ID, row number, or a Row object.
        :param column: The column ID or title.

        Examples:

        *Robot Framework:*

        .. code-block:: robotframework

            ${cell_history}=  Get Cell History  1  Approval
            FOR  ${revision} IN  @{cell_history}
                Log  Modified by ${revision.modified_by.email}
            END

        *Python:*

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            cell_history = ss.get_cell_history(1, "Approval")
            for revision in cell_history:
                print(f"Modified by {revision.modified_by.email}")
        """
        self._require_current_sheet()
        row_id = self._parse_row_id(row)
        column_id = self._get_column_id(column)
        cell_history = self.smart.Cells.get_cell_history(
            self.current_sheet.id, row_id, column_id, include_all=True
        )
        return self._unpack_index_result(cell_history)

    # *** ATTACHMENTS ***
    def _download_row_attachments(
        self, row: Union[int, Row], download_path: PathType = None
    ) -> List[Path]:
        """Downloads all attachments from a row, saves them locally and
        returns a list of Paths to those attachements. Defaults to
        saving in the robot's ``OUTPUT_DIR``, but can be specified.
        Requires a current sheet.

        If ``download_path`` is not set, it will attempt to use the
        attribute ``download_path`` set to the ``current_sheet``.

        :param row: The row ID, row number, or a Row object.
        :param download_path: The path to save the attachments to.
        """
        self._require_current_sheet()
        attachment_result = self.smart.Attachments.list_row_attachments(
            self.current_sheet.id, self._parse_row_id(row), include_all=True
        )
        attachments = self._unpack_index_result(attachment_result)
        download_path = Path(
            download_path
            or getattr(self.current_sheet, "download_path", None)
            or get_output_dir()
        )
        return [self.download_attachment(a, download_path) for a in attachments]

    def _download_attachment_by_id(
        self, attachment_id: int, download_path: Path = None
    ) -> Path:
        """Downloads the provided attachment using it's ID, requires
        a current sheet be selected. If ``download_path`` is not set,
        it will attempt to use the attribute ``download_path`` set to
        the ``current_sheet``.
        """
        self._require_current_sheet()
        attachment = self.smart.Attachments.get_attachment(
            self.current_sheet.id, attachment_id
        )
        download_result = self.smart.Attachments.download_attachment(
            attachment,
            str(
                download_path
                or getattr(self.current_sheet, "download_path", None)
                or get_output_dir()
            ),
        )
        return Path(download_result.download_directory) / download_result.filename

    @keyword
    def list_attachments(self) -> Optional[List[Attachment]]:
        """Gets a list of all attachments from the currently selected sheet.

        This will include attachments to the sheet, rows, and discussions.

        Examples:

        *Robot Framework:*

        .. code-block:: robotframework

            ${attachments}=  List Attachments
            FOR  ${attachment} IN  @{attachments}
                Log  ${attachment.name}
            END

        *Python:*

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            attachments = ss.list_attachments()
            for attachment in attachments:
                print(attachment.name)
        """
        self._require_current_sheet()
        attachment_result = self.smart.Attachments.list_all_attachments(
            self.current_sheet.id, include_all=True
        )
        return self._unpack_index_result(attachment_result)

    @keyword
    @_require_auth
    def download_attachment(
        self,
        attachment: Union[int, str, Dict, Attachment],
        download_path: Optional[Union[str, Path]] = None,
    ) -> Path:
        """Downloads the provided attachment from the currently selected
        sheet to the provided download_path, which defaults to
        the ``${OUTPUT_DIR}``.

        The attachment can be provided as an integer representing the
        attachments ID, a dictionary with at least the key ``id`` or as
        the native ``Attachment`` data model type.

        :param attachment: An integar representing the attachment ID, a
         dictionary with at least the key ``id``, or a native
         ``Attachment`` data model object.
        :param download_path: The path to save the attachment to.

        Examples:

        *Robot Framework:*

        .. code-block:: robotframework

            ${attachment}=  Get Attachment  123456789
            ${path}=  Download Attachment  ${attachment}
            Log  ${path}

        *Python:*

        .. code-block:: python

            ss = Smartsheet(access_token=access_token)
            attachment = ss.get_attachment(123456789)
            path = ss.download_attachment(attachment)
            print(path)
        """
        download_path = Path(download_path or get_output_dir()).expanduser().resolve()
        download_path.mkdir(parents=True, exist_ok=True)

        if isinstance(attachment, (int, str)):
            try:
                return self._download_attachment_by_id(attachment, download_path)
            except TypeError as e:
                raise TypeError("Attachment ID must be an integer.") from e
        elif isinstance(attachment, dict):
            if attachment.get("id", False):
                return self._download_attachment_by_id(attachment["id"], download_path)
            else:
                raise ValueError(
                    "The provided attachment dictionary does not "
                    "contain the 'id' key."
                )
        elif isinstance(attachment, Attachment):
            try:
                return self._download_attachment_by_id(attachment.id, download_path)
            except AttributeError as e:
                raise AttributeError("Attachment object does not have an ID.") from e
        else:
            raise TypeError(
                f"Invalid attachment type: received '{type(attachment)}', "
                f"expected 'int', 'str', 'dict' or 'Attachment'."
            )

    # *** USERS ***
    @keyword
    @_require_auth
    def get_current_user(self) -> User:
        """Gets the current authenticated user, which is also set in
        the library's memory as the current user. Call this again
        if you switch user or begin to impersonate a user.
        """
        self.current_user = self.smart.Users.get_current_user()
        return self.current_user



# pylint: disable=too-many-lines
import base64
import logging
import mimetypes
import os
import re
import time

# email package declares these properties in the __all__ definition but
# pylint ignores that
from email import (
    encoders,
    message_from_bytes,
    message_from_binary_file,
)  # pylint: disable=E0611
from email.charset import QP, add_charset  # pylint: disable=E0611
from email.generator import Generator  # pylint: disable=E0611
from email.header import decode_header, make_header  # pylint: disable=E0611
from email.message import Message  # pylint: disable=E0611
from email.mime.base import MIMEBase  # pylint: disable=E0611
from email.mime.image import MIMEImage  # pylint: disable=E0611
from email.mime.multipart import MIMEMultipart  # pylint: disable=E0611
from email.mime.text import MIMEText  # pylint: disable=E0611
from enum import Enum
from functools import wraps
from imaplib import IMAP4_SSL
from io import StringIO
from pathlib import Path
from smtplib import (
    SMTP,
    SMTP_SSL,
    SMTPConnectError,
    SMTPNotSupportedError,
    SMTPServerDisconnected,
    ssl,
)
from typing import Any, BinaryIO, List, Optional, Tuple, Union, Dict

from RPA.Email.common import (
    OAuthMixin,
    OAuthProvider,
    OAuthProviderType,
    counter_duplicate_path,
    NoRecipientsError,
)
from RPA.Robocorp.utils import protect_keywords


FilePath = Union[str, Path]

FLAG_DELETED = "\\Deleted"
FLAG_SEEN = "\\Seen"
FLAG_FLAGGED = "\\Flagged"
FLAG_TRASH = "\\Trash"


class AttachmentPosition(Enum):
    """Possible attachment positions in the message content."""

    TOP = 1  # Default
    BOTTOM = 2


def to_list(value):
    if value is None:
        return []
    if not isinstance(value, list):
        return value.split(",")
    return value


def to_attachment_position(value):
    """Convert value to AttachmentPosition enum."""
    if isinstance(value, AttachmentPosition):
        return value

    sanitized = str(value).upper().strip().replace(" ", "_")
    try:
        return AttachmentPosition[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown AttachmentPosition: {value}") from err


class Action(Enum):
    """Possible email actions."""

    msg_copy = 1
    msg_delete = 2
    msg_list = 3
    msg_flag = 4
    msg_unflag = 5
    msg_read = 6
    msg_unread = 7
    msg_save = 8
    msg_attachment_save = 9
    msg_move = 10
    msg_trash = 11
    glabel_add = 20  # Add GMail label
    glabel_remove = 21  # Remove GMail label


def to_action(value):
    """Convert value to Action enum."""
    if isinstance(value, Action):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return Action[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown email action: {value}") from err


def get_payload_decoded_or_not(message_part):
    result = None
    try:
        result = message_part.get_payload(decode=True).decode("utf-8")
    except Exception:  # pylint: disable=W0703
        result = message_part.get_payload(decode=True)
    return result


def get_email_body_from_message(message: Message, message_dict: Dict):
    if message.is_multipart():
        for part in message.walk():
            if part.get_content_type() == "text/plain":
                message_dict["body_text"] = get_payload_decoded_or_not(part)
            elif part.get_content_type() == "text/html":
                message_dict["body_html"] = get_payload_decoded_or_not(part)
    else:
        message_dict["body_text"] = get_payload_decoded_or_not(message)


def get_attachments_from_message(
    message: Message, save_attachments_directory: str, message_dict: Dict
):
    attachments = []
    if message.is_multipart():
        for part in message.walk():
            if part.get_content_maintype() == "multipart":
                continue
            if part.get("Content-Disposition") is None:
                continue

            filename = part.get_filename()
            if not filename:
                ext = part.get_content_type().split("/")[-1]
                filename = f"untitled.{ext}"

            if save_attachments_directory:
                data = part.get_payload(decode=True)
                full_filepath = os.path.join(save_attachments_directory, filename)
                with open(full_filepath, "wb") as f:
                    f.write(data)
                attachments.append(full_filepath)
            else:
                attachments.append(filename)

    message_dict["attachments"] = attachments


def get_part_filename(msg: Message) -> Optional[str]:
    filename = msg.get_filename()
    if not filename:
        return None

    decoded = decode_header(filename)
    if decoded[0][1] is not None:
        filename = decoded[0][0].decode(decoded[0][1])
    return filename.replace("\r", "").replace("\n", "")


def imap_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].imap_conn is None:
            raise ValueError("Requires authorized IMAP connection")
        return f(*args, **kwargs)

    return wrapper


def smtp_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].smtp_conn is None:
            raise ValueError("Requires authorized SMTP connection")
        return f(*args, **kwargs)

    return wrapper


class ImapSmtp(OAuthMixin):
    """`ImapSmtp` is a library for sending, reading, and deleting emails.
    `ImapSmtp` is interfacing with SMTP and IMAP protocols.

    ***About criteria argument***

    Various keywords like ``List Messages`` and ``Move Messages`` have keyword
    argument called ``criterion`` which can be used to filter emails according
    to given criteria.

    Syntax needs to according to specification and more information about that
    can be read from https://robocorp.com/docs/development-guide/email/sending-emails-with-gmail-smtp#listing-email-messages-by-criteria

    **Troubleshooting**

    - Authentication error with GMail - "Application-specific password required"
        See: https://support.google.com/mail/answer/185833?hl=en
    - More secure apps (XOAUTH2 protocol) - Use the OAuth2 flow as in this Portal robot:
        `example-oauth-email <https://github.com/robocorp/example-oauth-email>`_

        Make sure to specify a `provider` (and optionally a `tenant`) when importing
        the library and planning to use this flow.

    **Examples**

    **Robot Framework**

    It is highly recommended to secure your passwords and take care
    that they are not stored in version control by mistake.
    See ``RPA.Robocorp.Vault`` to see how to store secrets in
    Robocorp Vault.

    When sending HTML content with IMG tags, the ``src`` filenames must match
    the base image name given with the ``images`` parameter.

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Email.ImapSmtp   smtp_server=smtp.gmail.com  smtp_port=587
        Task Setup  Authorize  account=${GMAIL_ACCOUNT}  password=${GMAIL_PASSWORD}

        *** Variables ***
        ${GMAIL_ACCOUNT}        ACCOUNT_NAME
        ${GMAIL_PASSWORD}       APP_PASSWORD
        ${RECIPIENT_ADDRESS}    RECIPIENT
        ${BODY_IMG1}            ${IMAGEDIR}${/}approved.png
        ${BODY_IMG2}            ${IMAGEDIR}${/}invoice.png
        ${EMAIL_BODY}     <h1>Heading</h1><p>Status: <img src='approved.png' alt='approved image'/></p>
        ...               <p>INVOICE: <img src='invoice.png' alt='invoice image'/></p>

        *** Tasks ***
        Sending email
            Send Message  sender=${GMAIL_ACCOUNT}
            ...           recipients=${RECIPIENT_ADDRESS}
            ...           subject=Message from RPA Robot
            ...           body=RPA Robot message body

        Sending HTML Email With Image
            [Documentation]     Sending email with HTML content and attachment
            Send Message
            ...                 sender=${GMAIL_ACCOUNT}
            ...                 recipients=${RECIPIENT_ADDRESS}
            ...                 subject=HTML email with body images (2) plus one attachment
            ...                 body=${EMAIL_BODY}
            ...                 html=${TRUE}
            ...                 images=${BODY_IMG1}, ${BODY_IMG2}
            ...                 attachments=example.png

    **Python**

    .. code-block:: python

        from RPA.Email.ImapSmtp import ImapSmtp

        gmail_account = "ACCOUNT_NAME"
        gmail_password = "APP_PASSWORD"

        mail = ImapSmtp(smtp_server="smtp.gmail.com", smtp_port=587)
        mail.authorize(account=gmail_account, password=gmail_password)
        mail.send_message(
            sender=gmail_account,
            recipients="RECIPIENT",
            subject="Message from RPA Python",
            body="RPA Python message body",
        )
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    TO_PROTECT = [
        "authorize",
        "authorize_imap",
        "authorize_smtp",
        "set_credentials",
        "generate_oauth_string",
    ] + OAuthMixin.TO_PROTECT

    def __init__(
        self,
        smtp_server: Optional[str] = None,
        smtp_port: int = 587,
        imap_server: Optional[str] = None,
        imap_port: int = 993,
        account: Optional[str] = None,
        password: Optional[str] = None,
        encoding: str = "utf-8",
        provider: OAuthProviderType = OAuthProvider.GOOGLE,
        tenant: Optional[str] = None,
    ) -> None:
        # Init the OAuth2 support. (ready if used)
        super().__init__(provider, tenant=tenant)

        protect_keywords("RPA.Email.ImapSmtp", self.TO_PROTECT)
        self.logger = logging.getLogger(__name__)

        self.smtp_server = smtp_server
        self.imap_server = imap_server if imap_server else smtp_server
        self.smtp_port = int(smtp_port)
        self.imap_port = int(imap_port)
        self.encoding = encoding
        self.account = self.password = None
        self.set_credentials(account, password)

        self.smtp_conn = None
        self.imap_conn = None
        self.selected_folder = None

    def __del__(self) -> None:
        if self.smtp_conn:
            try:
                self.smtp_conn.quit()
            except SMTPServerDisconnected:
                self.logger.debug("Was already disconnected")
            finally:
                self.smtp_conn = None

        if self.imap_conn:
            try:
                self.imap_conn.close()
                self.imap_conn.logout()
            except Exception:  # pylint: disable=W0703
                pass
            self.imap_conn = None

    def set_credentials(
        self, account: Optional[str] = None, password: Optional[str] = None
    ) -> None:
        """Set credentials

        :param account: user account as string, defaults to None
        :param password: user password as string, defaults to None

        Example:

        .. code-block:: robotframework

            Set Credentials   ${username}   ${password}
            Authorize
        """
        self.account = account
        self.password = password

    def authorize_smtp(
        self,
        account: str = None,
        password: str = None,
        smtp_server: str = None,
        smtp_port: int = None,
        is_oauth: bool = False,
    ) -> None:
        """Authorize to SMTP server.

        :param account: SMTP account name, defaults to None
        :param password: SMTP account password, defaults to None
        :param smtp_server: SMTP server address, defaults to None
        :param smtp_port: SMTP server port, defaults to None (587 for SMTP)
        :param is_oauth: Use XOAUTH2 protocol with a base64 encoded OAuth2 string as
            `password`

        Can be called without giving any parameters if library
        has been initialized with necessary information and/or
        keyword ``Set Credentials`` has been called.

        Example:

        .. code-block:: robotframework

            Authorize SMTP    ${username}   ${password}  smtp.gmail.com  587
        """
        account = account or self.account
        password = password or self.password
        smtp_server = smtp_server or self.smtp_server
        smtp_port = smtp_port or self.smtp_port
        if smtp_server:
            try:
                self.smtp_conn = SMTP(smtp_server, smtp_port)
                self.send_smtp_hello()
                try:
                    self.smtp_conn.starttls()
                except SMTPNotSupportedError:
                    self.logger.warning("TLS not supported by the server")
            except SMTPConnectError:
                context = ssl.create_default_context()
                self.smtp_conn = SMTP_SSL(smtp_server, smtp_port, context=context)
            if account and password:
                if is_oauth:
                    self.smtp_conn.auth(
                        "XOAUTH2", lambda: base64.b64decode(password.encode()).decode()
                    )
                else:
                    self.smtp_conn.login(account, password)
        else:
            self.logger.warning("SMTP server address is needed for authentication")
        if self.smtp_conn is None:
            self.logger.warning("Not able to establish SMTP connection")

    def authorize_imap(
        self,
        account: str = None,
        password: str = None,
        imap_server: str = None,
        imap_port: int = None,
        is_oauth: bool = False,
    ) -> None:
        """Authorize to IMAP server.

        :param account: IMAP account name, defaults to None
        :param password: IMAP account password, defaults to None
        :param imap_server: IMAP server address, defaults to None
        :param imap_port: IMAP server port, defaults to None
        :param is_oauth: Use XOAUTH2 protocol with a base64 encoded OAuth2 string as
            `password`

        Can be called without giving any parameters if library
        has been initialized with necessary information and/or
        keyword ``Set Credentials`` has been called.

        Example:

        .. code-block:: robotframework

            Authorize IMAP    ${username}   ${password}  imap.gmail.com  993
        """
        if account is None and password is None:
            account = self.account
            password = self.password
        if imap_server is None:
            imap_server = self.imap_server
        if imap_port is None:
            imap_port = self.imap_port
        else:
            imap_port = int(imap_port)
        if imap_server and account and password:
            self.imap_conn = IMAP4_SSL(imap_server, imap_port)
            if is_oauth:
                self.imap_conn.authenticate(
                    "XOAUTH2", lambda _: base64.b64decode(password.encode())
                )
            else:
                self.imap_conn.login(account, password)
            self.imap_conn.select("INBOX")
        else:
            self.logger.warning(
                "Server address, account and password are needed for "
                "authentication with IMAP"
            )
        if self.imap_conn is None:
            self.logger.warning("Not able to establish IMAP connection")

    def authorize(
        self,
        account: str = None,
        password: str = None,
        smtp_server: str = None,
        imap_server: str = None,
        smtp_port: int = None,
        imap_port: int = None,
        is_oauth: bool = False,
    ) -> None:
        # pylint: disable=C0301
        """Authorize user to SMTP and IMAP servers.

        :param account: user account as string, defaults to None
        :param password: user password as string, defaults to None
        :param smtp_server: SMTP server address, defaults to None
        :param imap_server: IMAP server address, defaults to None
        :param smtp_port: SMTP server port, defaults to None (587 for SMTP)
        :param imap_port: IMAP server port, defaults to None
        :param is_oauth: Use XOAUTH2 protocol with a base64 encoded OAuth2 string as
            `password`

        Will use separately set credentials or those given in keyword call.

        Example:

        .. code-block:: robotframework

            Authorize    ${username}   ${password}  smtp_server=smtp.gmail.com  smtp_port=587
        """  # noqa: E501
        self.authorize_smtp(
            account, password, smtp_server, smtp_port, is_oauth=is_oauth
        )
        self.authorize_imap(
            account, password, imap_server, imap_port, is_oauth=is_oauth
        )

    @smtp_connection
    def send_smtp_hello(self) -> None:
        """Send hello message to SMTP server.

        Required step when creating SMTP connection.
        """
        if self.smtp_conn:
            self.smtp_conn.ehlo()

    @smtp_connection
    def send_message(
        self,
        sender: str,
        recipients: Optional[Union[List[str], str]] = None,
        subject: str = "",
        body: str = "",
        attachments: Optional[Union[List[str], str]] = None,
        html: bool = False,
        images: Optional[Union[List[str], str]] = None,
        cc: Optional[Union[List[str], str]] = None,
        bcc: Optional[Union[List[str], str]] = None,
        attachment_position: Optional[AttachmentPosition] = AttachmentPosition.TOP,
        in_reply_to: Optional[str] = None,
        return_path: Optional[str] = None,
        reply_to: Optional[str] = None,
    ) -> bool:
        """Send SMTP email

        :param sender: who is sending, ie. 'from'
        :param recipients: who is receiving, ie. 'to'
        :param subject: mail subject field
        :param body: mail body content
        :param attachments: list of filepaths to attach
        :param html: if message content is in HTML, default `False`
        :param images: list of filepaths for inline images
        :param cc: list of email addresses for email 'cc' field
        :param bcc: list of email addresses for email 'bcc' field
        :param attachment_position: content position for attachment, default `top`
        :param in_reply_to: the 'Message ID' to which this message is in reply to,
         for example `<message_id_for_reply_to>`
        :param return_path: email address which should receive "bounce messages"
        :param reply_to: email address which should receive the reply

        **Valid sender values**

        - First Lastname <address@domain>
        - address@domain

        **About in_reply_to**

        In addition of setting `in_reply_to` parameter to match the 'Message ID'
        of the email this message is replying to, some email servers require that
        also subject of the original email is included in the reply email subject
        with `re: ` prefix, ie. "re: Why it is best to use Robocorp RPA" or with
        the original email subject.

        And please note that `in_reply_to` the 'Message ID' needs to contain `<` in
        the start of the ID and `>` at the end of the ID.

        **About return_path**

        Email servers tend to set 'Return-Path' of the email on their own so in
        some cases user given address won't work as the email server does not
        use the user set address. If possible, the email server's configuration
        can be changed to ensure that 'Return-Path' header is respected.

        Example:

        .. code-block:: robotframework

            Send Message  sender@domain.com  recipient@domain.com
            ...           cc=need_to_know@domain.com
            ...           bcc=hidden_copy@domain.com
            ...           subject=Greetings Software Robot Developer
            ...           body=${email_body}
            ...           attachments=${CURDIR}${/}report.pdf

            # Fixing attachments to the bottom of the content
            Send Message  sender@domain.com  recipient@domain.com
            ...           subject=Greetings Software Robot Developer
            ...           body=${email_body}
            ...           attachments=${CURDIR}${/}report.pdf
            ...           attachment_position=bottom
        """
        add_charset(self.encoding, self.encoding, QP, self.encoding)
        msg = MIMEMultipart()
        msg["Subject"] = subject
        msg["From"] = sender
        if reply_to:
            msg["Reply-To"] = reply_to
        # The following lines optimize handling the parameters
        msg_recipients, attachments, images = self._handle_message_parameters(
            msg, recipients, cc, bcc, attachments, images
        )

        evaluated_attachment_position = to_attachment_position(attachment_position)

        if evaluated_attachment_position == AttachmentPosition.TOP:
            self._add_attachments_to_msg(attachments, msg)

        if return_path:
            msg.add_header("Return-Path", return_path)
        if in_reply_to:
            msg.add_header("In-Reply-To", in_reply_to)

        self._add_message_content(html, images, body, msg)

        if evaluated_attachment_position == AttachmentPosition.BOTTOM:
            self._add_attachments_to_msg(attachments, msg)

        # Create a generator and flatten message object to 'file’
        str_io = StringIO()
        g = Generator(str_io, False)
        g.flatten(msg)

        try:
            if self.smtp_conn is None:
                self.authorize_smtp()
            self.smtp_conn.sendmail(
                from_addr=sender, to_addrs=msg_recipients, msg=str_io.getvalue()
            )
        except Exception as err:
            raise ValueError(f"Send Message failed: {err}") from err

        return True

    def _add_message_content(self, html, images, body, msg):
        if html:
            for im in images:
                im = im.strip()
                imname = Path(im).name
                body = body.replace(str(imname), f"cid:{imname}")
                with open(im, "rb") as f:
                    img = MIMEImage(f.read())
                    img.add_header("Content-ID", f"<{imname}>")
                    msg.attach(img)
            htmlpart = MIMEText(body, "html", self.encoding)
            msg.attach(htmlpart)
        else:
            textpart = MIMEText(body, "plain", self.encoding)
            msg.attach(textpart)
            for im in images:
                im = im.strip()
                imname = Path(im).name
                with open(im, "rb") as f:
                    img = MIMEImage(f.read())
                    msg.add_header(
                        "Content-Disposition",
                        f"inline; filename= {imname}",
                    )
                    msg.attach(img)

    def _handle_message_parameters(self, msg, recipients, cc, bcc, attachments, images):
        # Convert the parameters to lists using the helper function
        recipients = to_list(recipients)
        attachments = to_list(attachments)
        images = to_list(images)

        has_recipients = False

        if recipients:
            has_recipients = True
            msg.add_header("To", ",".join(recipients))
        if cc:
            has_recipients = True
            msg["Cc"] = ",".join(to_list(cc))
            recipients += to_list(cc)
        if bcc:
            has_recipients = True
            recipients += to_list(bcc)

        if not has_recipients:
            raise NoRecipientsError(
                "Message needs to have either 'recipients', 'cc' or 'bcc' for sending."
            )

        return recipients, attachments, images

    def _add_attachments_to_msg(self, attachments: list = None, msg=None):
        if len(attachments) > 0:
            for filename in attachments:
                if os.path.dirname(filename) == "":
                    filename = str(Path.cwd() / filename)
                with open(filename, "rb") as attachment:
                    ctype, encoding = mimetypes.guess_type(filename)
                    if ctype is None or encoding is not None:
                        ctype = "application/octet-stream"

                    maintype, subtype = ctype.split("/", 1)
                    self.logger.warning(f"MAIN: {maintype} SUB: {subtype}")
                    if maintype == "image":
                        # image attachment
                        part = MIMEImage(
                            attachment.read(),
                            name=Path(filename).name,
                            _subtype=subtype,
                        )
                    elif maintype == "text":
                        part = MIMEText(
                            attachment.read().decode("utf-8"), _subtype=subtype
                        )
                    else:
                        # attach other filetypes
                        part = MIMEBase("application", "octet-stream")
                        part.set_payload(attachment.read())
                        encoders.encode_base64(part)
                    part.add_header(
                        "Content-Disposition",
                        "attachment",
                        filename=Path(filename).name,
                    )
                    msg.attach(part)

    def get_decoded_email_body(
        self, message, html_first: bool = False
    ) -> Tuple[str, bool]:
        """Decodes email body and extracts its text/html content.

        Automatically detects character set if the header is not set.

        :param message: Raw 7-bit message body input e.g. from `imaplib`. Double
            encoded in quoted-printable and latin-1
        :param html_first: Prioritize html extraction over text when this is True
        :returns: Message body as unicode string and a boolean telling if the message
            has attachments
        """
        if not message.is_multipart():
            content_charset = message.get_content_charset()
            text = str(
                message.get_payload(decode=True),
                content_charset or self.encoding,
                "ignore",
            )
            return text.strip(), False

        text = html = None
        has_attachments = False

        for part in message.walk():
            if not part:
                continue
            content_filename = get_part_filename(part)
            if content_filename:
                has_attachments = True
                continue
            content_type = "text/plain"
            _data = ""
            content_charset = part.get_content_charset()
            if content_charset:
                content_type = part.get_content_type()
            payload = part.get_payload(decode=True)
            if payload:
                _data = str(payload, str(content_charset), "ignore")

            if content_type == "text/plain":
                text = _data
            elif content_type == "text/html":
                html = _data

        if html_first:
            data = html or text
        else:
            data = text or html
        data = data.strip() if data else ""
        return data, has_attachments

    @imap_connection
    def _do_actions_on_messages(
        self,
        criterion: Union[str, dict],
        actions: list,
        labels: str = None,
        source_folder: str = None,
        target_folder: str = None,
        limit: int = None,
        overwrite: bool = False,
        readonly: bool = False,
        prefix: str = "",
    ) -> list:
        selected_folder = source_folder or self.selected_folder
        folders = self.get_folder_list(subdirectory=selected_folder)
        result = {"actions_done": 0, "message_count": 0, "ids": [], "uids": {}}

        if source_folder:
            for folder in folders:
                if "Noselect" in folder["flags"]:
                    continue
                folder_name = folder["name"]
                self.select_folder(folder_name, readonly)
                self._search_message(criterion, actions, limit, result)
        else:
            self._search_message(criterion, actions, limit, result)

        if len(result["uids"].keys()) == 0:
            self.logger.warning(
                "Can't find any messages matching criterion '%s' in source folder '%s'",
                criterion,
                source_folder or "INBOX",
            )
            return result
        if limit is None or len(result["uids"].keys()) <= limit:
            for mail in result["uids"].items():
                status = self._perform_actions(
                    actions,
                    mail,
                    labels=labels,
                    target_folder=target_folder,
                    overwrite=overwrite,
                    prefix=prefix,
                )
                if status:
                    result["actions_done"] += 1
        else:
            self.logger.warning(
                "Some emails were not processed because limit was reached"
            )
        return result

    def _search_message(self, criterion, actions, limit, result):
        search_encoding = None
        search_command = None
        literal_search = criterion.startswith("literal:")
        gmail_search = criterion.startswith("gmail:")

        if literal_search:
            search_encoding = "utf-8"
            search_term = criterion.replace("literal:", "")
            search_command, search_literal = search_term.split(" ", 1)

            self.imap_conn.literal = b"%s" % search_literal.encode("utf-8")
        elif gmail_search:
            search_encoding = "utf-8"
            search_command = "X-GM-RAW"
            self.imap_conn.literal = b"(%s)" % criterion.replace("gmail:", "").encode(
                "utf-8"
            )
        else:
            search_command = "(%s)" % criterion
        try:
            self.logger.info("IMAP search: '%s'", criterion)
            status, data = self.imap_conn.search(search_encoding, search_command)
        except Exception as err:  # pylint: disable=broad-except
            self.logger.warning(
                "Email search returned: %s (%s)", str(err), search_command
            )
            return False
        if status == "OK":
            mail_id_data = bytes.decode(data[0])
            mail_ids = mail_id_data.split() if len(mail_id_data) > 0 else []
            if limit and len(mail_ids) + result["message_count"] > limit:
                return True
            if len(mail_ids) > 0:
                result["message_count"] += len(mail_ids)
                result["ids"].extend(mail_ids)
                for mail_id in mail_ids:
                    mail_uid, message = self._fetch_uid_and_message(mail_id, actions)
                    if mail_uid is None or mail_uid in result["uids"].keys():
                        continue
                    message["uid"] = mail_uid
                    result["uids"][mail_uid] = message
            return True
        else:
            self.logger.warning("Search result not OK: %s", status)
            return False

    def _get_mail_uids_and_dicts(self, mail):
        mail_uids = []
        mail_dicts = []
        if isinstance(mail, list):
            mail_uids = [m["uid"] for m in mail]
            mail_dict = None
            mail_dicts = mail
        else:
            try:
                mail_uid, mail_dict = mail
                mail_uids = [mail_uid]
                mail_dicts = [mail_dict]
            except ValueError:
                mail_uid = mail["uid"]
                mail_dict = mail
                mail_uids = [mail_uid]
                mail_dicts = [mail]
        return mail_uids, mail_dicts

    def _command_mail_uid(self, command, uid, m_op="", m_param="", target_folder=None):
        if target_folder:
            action_status, _ = self.imap_conn.uid(command, uid, target_folder)
            self.logger.debug(
                "%s %s %s RESULT %s", command, uid, target_folder, action_status
            )
        else:
            action_status, _ = self.imap_conn.uid(command, uid, m_op, m_param)
            self.logger.info(
                "%s %s %s %s RESULT %s",
                command,
                uid,
                m_op,
                m_param,
                action_status,
            )
        return action_status

    def _perform_actions(  # noqa: C901 pylint: disable=too-many-branches
        self,
        actions: list,
        mail: Union[dict, list],
        labels: str = None,
        target_folder: str = None,
        overwrite: bool = False,
        prefix: str = "",
    ):
        action_status = True
        mail_uids, mail_dicts = self._get_mail_uids_and_dicts(mail)
        store_params = {
            Action.glabel_add: ["+X-GM-LABELS", f"({labels})"],
            Action.glabel_remove: ["-X-GM-LABELS", f"({labels})"],
            Action.msg_delete: ["+FLAGS", FLAG_DELETED],
            Action.msg_flag: ["+FLAGS", FLAG_FLAGGED],
            Action.msg_unflag: ["-FLAGS", FLAG_FLAGGED],
            Action.msg_read: ["+FLAGS", FLAG_SEEN],
            Action.msg_unread: ["-FLAGS", FLAG_SEEN],
            Action.msg_trash: ["+FLAGS", FLAG_TRASH],
        }
        for act in actions:
            action = to_action(act)
            if action == Action.msg_list:
                continue
            if action in store_params.keys():
                m_op, m_param = store_params[action]
                for uid in mail_uids:
                    action_status = self._command_mail_uid("STORE", uid, m_op, m_param)
            elif action == Action.msg_copy:
                for uid in mail_uids:
                    action_status = self._command_mail_uid(
                        "COPY", uid, target_folder=target_folder
                    )
            elif action == Action.msg_move:
                m_op, m_param = store_params[Action.msg_delete]
                for uid in mail_uids:
                    if " " in target_folder:
                        target_folder = f'"{target_folder}"'
                    action_status = self._command_mail_uid(
                        "COPY", uid, target_folder=target_folder
                    )
                    action_status = self._command_mail_uid("STORE", uid, m_op, m_param)
            elif action == Action.msg_save and len(mail_dicts) > 0:
                if target_folder is None:
                    target_folder = os.path.expanduser("~")
                for mail_dict in mail_dicts:
                    action_status = self._save_eml_file(
                        mail_dict, target_folder, overwrite, prefix
                    )
            elif action == Action.msg_attachment_save and len(mail_dicts) > 0:
                if target_folder is None:
                    target_folder = os.path.expanduser("~")
                for mail_dict in mail_dicts:
                    self._save_attachment(mail_dict, target_folder, overwrite, prefix)
            else:
                # TODO: mypy should handle enum exhaustivity validation
                raise ValueError(
                    f"Unsupported email action or insufficient input data: {action}"
                )
        return action_status

    @imap_connection
    def delete_message(
        self,
        criterion: Union[str, dict] = None,
        source_folder: str = None,
    ) -> bool:
        """Delete single message from server based on criterion.

        :param criterion: filter messages based on this search, can also be a
         message dictionary
        :param source_folder: defaults to already selected folder, but can be
         set to delete message in a specific folder
        :return: True if success, False if not

        If criterion does not return exactly 1 message then delete is not done.

        Example:

        .. code-block:: robotframework

            Delete Message  SUBJECT \"Greetings RPA developer\"
        """
        selected_folder = source_folder or self.selected_folder
        if isinstance(criterion, dict):
            readonly = False
            self.select_folder(selected_folder, readonly)
            status = self._perform_actions(
                [Action.msg_delete],
                criterion,
            )
            result = {
                "actions_done": 1 if status else 0,
                "message_count": 1,
                "ids": [criterion["uid"]],
                "uids": {criterion["uid"]: criterion},
            }
            self.imap_conn.expunge()
        else:
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_delete],
                limit=1,
                source_folder=selected_folder,
            )
            self.imap_conn.expunge()
        return result["actions_done"] == 1

    @imap_connection
    def delete_messages(
        self,
        criterion: Union[str, list] = None,
        limit: int = None,
        source_folder: str = None,
    ) -> bool:
        """Delete messages from server based on criterion.

        :param criterion: filter messages based on this search, can also be a
         list of message dictionaries
        :param limit: maximum number of message to delete
        :param source_folder: defaults to already selected folder, but can be
         set to delete message in a specific folder
        :return: True if success, False if not

        Example:

        .. code-block:: robotframework

            Delete Messages  SUBJECT Greetings
        """
        selected_folder = source_folder or self.selected_folder
        if isinstance(criterion, list):
            readonly = False
            self.select_folder(selected_folder, readonly)
            status = self._perform_actions(
                [Action.msg_delete],
                criterion,
            )
            self.imap_conn.expunge()
            return status
        else:
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_delete],
                limit=limit,
                source_folder=selected_folder,
            )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @imap_connection
    def save_messages(
        self,
        criterion: Optional[Union[str, dict, list]] = None,
        target_folder: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> bool:
        # pylint: disable=C0301
        """Save messages based on criteria and store them to target folder
        with attachment files.

        Does not save message if `target_folder` is not given.

        :param criterion: filter messages based on this, defaults to ""
        :param target_folder: path to folder where message are saved, defaults to None
        :param prefix: optional filename prefix added to the attachments, empty by
            default
        :return: True if succeeded, False otherwise

        Example:

        .. code-block:: robotframework

            Save Messages  SUBJECT Important message  target_folder=${USERDIR}${/}messages
        """  # noqa: E501
        prefix = prefix or ""
        if target_folder is None:
            target_folder = os.path.expanduser("~")
        if isinstance(criterion, (dict, list)):
            status = self._perform_actions(
                [Action.msg_save], criterion, target_folder=target_folder, prefix=prefix
            )
            return status
        else:
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_save],
                target_folder=target_folder,
                prefix=prefix,
            )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @imap_connection
    def list_messages(
        self, criterion: str = "", source_folder: str = None, readonly: bool = True
    ) -> Any:
        """Return list of messages matching criterion.

        :param criterion: list emails matching this, defaults to ""
        :param source_folder: list messages from this folder
        :param readonly: set False if you want to mark matching messages as read
        :return: list of messages

        *Note.* listing messages without `source_folder` might take a long time

        Example:

        .. code-block:: robotframework

            @{emails}  List Messages  SUBJECT "rpa task"
            FOR  ${email}  IN  @{EMAILS}
                Log  ${email}[Subject]
                Log  ${email}[From]
                Log  ${email}[Date]
                Log  ${email}[Delivered-To]
                Log  ${email}[Received]
                Log  ${email}[Has-Attachments]
                Log  ${email}[uid]
            END
        """
        result = self._do_actions_on_messages(
            criterion,
            source_folder=source_folder,
            actions=[Action.msg_list],
            readonly=readonly,
        )
        values = result["uids"].values()
        converted = []
        for v in values:
            if v:
                converted.append(
                    {
                        str(key): (
                            value
                            if isinstance(value, (str, bool, int, Message))
                            else str(value)
                        )
                        for key, value in v.items()
                    }
                )
        return converted

    @imap_connection
    def save_attachments(
        self,
        criterion: str = "",
        target_folder: Optional[str] = None,
        overwrite: bool = False,
        prefix: Optional[str] = None,
    ) -> List[str]:
        # pylint: disable=C0301
        """Save mail attachments of emails matching criterion on the local disk.

        :param criterion: attachments are saved for mails matching this, defaults to ""
        :param target_folder: local folder for saving attachments to (needs to exist),
            defaults to user's home directory if None
        :param overwrite: overwrite existing file if True, defaults to False
        :param prefix: optional filename prefix added to the attachments, empty by
            default
        :return: list of saved attachments (absolute file paths) of all emails

        Example:

        .. code-block:: robotframework

            ${attachments} =    Save Attachments    SUBJECT "rpa task"
            ...    target_folder=${CURDIR}${/}messages  overwrite=${True}
            FOR  ${file}  IN  @{attachments}
                OperatingSystem.File Should Exist  ${file}
            END
        """  # noqa: E501
        attachments_saved = []
        prefix = prefix or ""
        messages = self.list_messages(criterion)
        for msg in messages:
            attachments_saved.extend(
                self.save_attachment(msg, target_folder, overwrite, prefix)
            )
        return attachments_saved

    def save_attachment(
        self,
        message: Union[dict, Message],
        target_folder: Optional[str],
        overwrite: bool,
        prefix: Optional[str] = None,
    ) -> List[str]:
        # pylint: disable=C0301
        """Save mail attachment of a single given email on the local disk.

        :param message: message item
        :param target_folder: local folder for saving attachments to (needs to exist),
            defaults to user's home directory if None
        :param overwrite: overwrite existing file if True, defaults to False
        :param prefix: optional filename prefix added to the attachments, empty by
            default
        :return: list of saved attachments (list of absolute filepaths) in one email

        Example:

        .. code-block:: robotframework

            @{emails} =    List Messages    ALL
            FOR    ${email}    IN    @{emails}
                IF    ${email}[Has-Attachments]
                    Log To Console    Saving attachment for: ${email}[Subject]
                    ${attachments} =    Save Attachment
                    ...    ${email}
                    ...    target_folder=${CURDIR}
                    ...    overwrite=${True}
                    Log To Console    Saved attachments: ${attachments}
                END
            END
        """  # noqa: E501
        prefix = prefix or ""
        if target_folder is None:
            target_folder = os.path.expanduser("~")
        return self._save_attachment(message, target_folder, overwrite, prefix)

    def _save_attachment(self, message, target_folder, overwrite, prefix) -> List[str]:
        attachments_saved = []
        msg = message["Message"] if isinstance(message, dict) else message

        for part in msg.walk():
            content_maintype = part.get_content_maintype()
            filename = None
            if content_maintype != "multipart":
                filename = get_part_filename(part)
            if not filename:
                continue

            filepath = Path(target_folder) / Path(f"{prefix}{filename}").name
            if not overwrite:
                filepath = counter_duplicate_path(filepath)
            self.logger.info("Attachment filepath: %r", filepath)
            payload = part.get_payload(decode=True)
            if payload:
                self.logger.info(
                    "Saving attachment: %s",
                    filename,
                )
                with open(filepath, "wb") as f:
                    f.write(payload)
                    attachments_saved.append(str(filepath))
            else:
                self.logger.info(
                    "Attachment %r did not have payload to write",
                    filename,
                )
        return attachments_saved

    def _save_eml_file(self, message, target_folder, overwrite, prefix):
        save_status = True
        emlfile = f"{prefix}{message['uid']}.eml"
        full_emlfile_path = Path(os.path.join(target_folder, emlfile))
        if not full_emlfile_path.exists() or overwrite:
            with open(full_emlfile_path, "wb") as f:
                if "bytes" in message.keys():
                    f.write(message["bytes"])
                elif "Message" in message:
                    f.write(message["Message"].as_bytes())
                else:
                    save_status = False
                    self.logger.error(
                        "Save failed. Unable to get message byte content."
                    )
        elif full_emlfile_path.exists() and not overwrite:
            self.logger.warning("Did not overwrite file: %s", full_emlfile_path)
        return save_status

    @imap_connection
    def wait_for_message(
        self,
        criterion: str = "",
        timeout: float = 5.0,
        interval: float = 1.0,
        readonly: bool = True,
    ) -> Any:
        """Wait for email matching `criterion` to arrive into mailbox.

        :param criterion: message filter to wait for, defaults to ""
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :param readonly: set False if you want to mark matching messages as read
        :return: list of messages

        Example:

        .. code-block:: robotframework

            @{emails}  Wait For Message  SUBJECT \"rpa task\"  timeout=300  interval=10
        """
        self._validate_criterion(criterion)
        end_time = time.time() + float(timeout)
        while time.time() < end_time:
            self.imap_conn.select("inbox", readonly)
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_list],
                source_folder="INBOX",
                readonly=readonly,
            )
            if result["message_count"] > 0:
                return result["uids"].values()
            time.sleep(interval)
        return []

    def _parse_folders(self, folders):
        if folders and len(folders) == 1 and folders[0] is None:
            return []
        parsed_folders = []
        folder_regex = r'\((?P<flags>.*)\)\s"(?P<delimiter>.*)"\s"?(?P<name>[^"]*)"?'
        for f in folders:
            match = re.search(folder_regex, bytes.decode(f))
            if not match:
                self.logger.warning("Cannot parse folder name %s", bytes.decode(f))
                continue
            parsed_folders.append(match.groupdict())
        return parsed_folders

    @imap_connection
    def get_folder_list(self, subdirectory: str = None, pattern: str = None) -> list:
        """Get list of folders on the server

        :param subdirectory: list subdirectories for this folder
        :param pattern: list folders matching this pattern
        :return: list of folders

        Example:

        .. code-block:: robotframework

            @{folders}  Get Folder List
            @{folders}  Get Folder List  pattern=important
            @{folders}  Get Folder List  subdirectory=sub
        """
        kwparams = {}
        if subdirectory:
            kwparams["directory"] = f'"{subdirectory}"'
        if pattern:
            kwparams["pattern"] = pattern

        status_code, folders = self.imap_conn.list(**kwparams)

        if status_code == "OK":
            parsed_folders = self._parse_folders(folders)
            return [
                f
                for f in parsed_folders
                if subdirectory is None or subdirectory == f["name"]
            ]
        else:
            return []

    @imap_connection
    def select_folder(self, folder_name: str = "INBOX", readonly: bool = False) -> int:
        """Select folder by name

        :param folder_name: name of the folder to select
        :param readonly: if set to True then message flags are not modified
        :return: message count in the selected folder

        Returns number of messages in the folder or
        exception if folder does not exist on the server.

        Example:

        .. code-block:: robotframework

            Select Folder   subfolder
        """
        status_code, data = self.imap_conn.select(
            mailbox=f'"{folder_name}"', readonly=readonly
        )
        if status_code == "OK":
            message_count = bytes.decode(data[0])
            self.selected_folder = folder_name
            return int(message_count)
        else:
            raise ValueError("Folder '%s' does not exist on the server" % folder_name)

    @imap_connection
    def rename_folder(
        self, oldname: str = None, newname: str = None, suppress_error: bool = False
    ) -> bool:
        """Rename email folder

        :param oldname: current folder name
        :param newname: new name for the folder
        :param suppress_error: to silence warning message, defaults to False
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Rename Folder   subfolder   filtered
        """
        if oldname is None or newname is None:
            raise KeyError(
                "Both 'oldname' and 'newname' and required for rename folder"
            )
        self.logger.info("Rename folder '%s' to '%s'", oldname, newname)
        status_code, data = self.imap_conn.rename(oldname, newname)

        if status_code == "OK":
            return True
        else:
            if suppress_error is False:
                self.logger.warning(
                    "Folder rename failed with message: '%s'",
                    bytes.decode(data[0]),
                )
            return False

    @imap_connection
    def delete_folder(self, folder_name: str = None) -> bool:
        """Delete email folder

        :param folder_name: current folder name
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Delete Folder   filtered
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for delete folder")
        self.logger.info("Delete folder '%s'", folder_name)
        status_code, data = self.imap_conn.delete(folder_name)
        if status_code == "OK":
            return True
        else:
            self.logger.warning(
                "Delete folder '%s' response status: '%s %s'",
                folder_name,
                status_code,
                bytes.decode(data[0]),
            )
            return False

    @imap_connection
    def create_folder(self, folder_name: str = None) -> bool:
        """Create email folder

        :param folder_name: name for the new folder
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Create Folder   filtered
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for create folder")
        self.logger.info("Create folder '%s'", folder_name)
        status_code, data = self.imap_conn.create(folder_name)
        if status_code == "OK":
            return True
        else:
            self.logger.warning(
                "Create folder '%s' response status: '%s %s'",
                folder_name,
                status_code,
                bytes.decode(data[0]),
            )
            return False

    @imap_connection
    def flag_messages(
        self, criterion: Union[str, dict] = None, unflag: bool = False
    ) -> Any:
        """Mark messages as `flagged`

        :param criterion: mark messages matching criterion
        :param unflag: to mark messages as not `flagged`
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${flagged}  ${oftotal}    Flag Messages   SUBJECT rpa
            ${unflagged}  ${oftotal}  Flag Messages   SUBJECT rpa  unflag=True
        """
        action = Action.msg_unflag if unflag else Action.msg_flag
        if isinstance(criterion, dict):
            status = self._perform_actions(mail=criterion, actions=[action])
            return status
        else:
            result = self._do_actions_on_messages(criterion, actions=[action])
            return (
                result["actions_done"] > 0
                and result["actions_done"] == result["message_count"]
            )

    @imap_connection
    def unflag_messages(self, criterion: Union[str, dict] = None) -> Any:
        """Mark messages as not `flagged`

        :param criterion: mark messages matching criterion
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${unflagged}  ${oftotal}  Unflag Messages   SUBJECT rpa
        """
        return self.flag_messages(criterion, unflag=True)

    @imap_connection
    def mark_as_read(
        self, criterion: Union[str, dict] = None, unread: bool = False
    ) -> Any:
        """Mark messages as `read`

        :param criterion: mark messages matching criterion
        :param unread: to mark messages as not `read`
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${read}  ${oftotal}  Mark As Read   SUBJECT rpa
        """
        action = Action.msg_unread if unread else Action.msg_read
        if isinstance(criterion, dict):
            status = self._perform_actions(mail=criterion, actions=[action])
            return status
        else:
            result = self._do_actions_on_messages(criterion, actions=[action])
            return (
                result["actions_done"] > 0
                and result["actions_done"] == result["message_count"]
            )

    @imap_connection
    def mark_as_unread(self, criterion: Union[str, dict] = None) -> Any:
        """Mark messages as not `read`

        :param criterion: mark messages matching criterion
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${unread}  ${oftotal}  Mark As Unread   SUBJECT rpa
        """
        return self.mark_as_read(criterion, unread=True)

    def _validate_criterion(self, criterion: str) -> bool:
        if criterion is None or len(criterion) < 1:
            raise KeyError("Criterion is required parameter")

    def _fetch_uid_and_message(self, mail_id, actions):
        message_dict = None
        _, data = self.imap_conn.fetch(mail_id, "(UID RFC822)")
        pattern_uid = re.compile(r".*UID (\d+) RFC822")
        decoded_data = bytes.decode(data[0][0]) if data[0] else None
        self.logger.debug("message identification: %s", decoded_data)
        match_result = pattern_uid.match(decoded_data) if decoded_data else None
        uid = match_result.group(1) if match_result else None
        if uid:
            message_dict = self._fetch_message_dict(mail_id, data, actions)

        key_to_change = None
        message_id_to_add = None
        if message_dict:
            for key, val in message_dict.items():
                if key.lower() == "message-id":
                    key_to_change = key
                    message_id_to_add = (
                        str(val).replace("<", "").replace(">", "").strip()
                    )
            if key_to_change:
                message_dict["Message-ID"] = message_id_to_add
                if key_to_change != "Message-ID":
                    del message_dict[key_to_change]
        return uid, message_dict

    def _fetch_message_dict(self, mail_id, data, actions):
        # _, data = self.imap_conn.fetch(mail_id, "(RFC822)")
        message = message_from_bytes(data[0][1])
        message_dict = {"Mail-Id": mail_id, "Message": message, "Body": ""}
        if Action.msg_save in actions:
            message_dict["bytes"] = data[0][1]
        for k, v in message.items():
            msg_item = decode_header(v)
            message_dict[k] = make_header(msg_item)
        (
            message_dict["Body"],
            has_attachments,
        ) = self.get_decoded_email_body(message)
        # SET DEFAULT VALUES FOR KEYS
        if "Delivered-To" not in message_dict.keys():
            message_dict["Delivered-To"] = ""
        message_dict["Has-Attachments"] = has_attachments
        return message_dict

    @imap_connection
    def move_messages(
        self,
        criterion: Union[str, dict] = None,
        target_folder: str = None,
        source_folder: str = None,
    ) -> bool:
        """Move messages from source folder to target folder

        :param criterion: move messages matching criterion
        :param source_folder: location of the messages, default `INBOX`
        :param target_folder: where messages should be move into
        :return: True if all move operations succeeded, False if not

        Example:

        .. code-block:: robotframework

            ${result}=    Move Messages
            ...    criterion=SUBJECT "order confirmation 32"
            ...    target_folder=yyy

            ${result}=    Move Messages
            ...    criterion=ALL
            ...    source_folder=yyy
            ...    target_folder=XXX
        """
        selected_folder = source_folder or self.selected_folder
        if target_folder is None or len(target_folder) == 0:
            raise KeyError("Can't move messages without target_folder")
        if isinstance(criterion, dict):
            readonly = False
            self.select_folder(selected_folder, readonly)
            status = self._perform_actions(
                mail=criterion,
                actions=[Action.msg_move],
                target_folder=target_folder,
            )
            result = {
                "actions_done": 1 if status else 0,
                "message_count": 1,
                "ids": [criterion["uid"]],
                "uids": {criterion["uid"]: criterion},
            }
        else:
            result = self._do_actions_on_messages(
                criterion=criterion,
                actions=[Action.msg_move],
                source_folder=source_folder,
                target_folder=target_folder,
            )
        if result["actions_done"] > 0:
            self.imap_conn.expunge()
        action_mismatch = result["actions_done"] != result["message_count"]
        if action_mismatch:
            self.logger.warning(
                "Criterion matched %s messages, but actions done to %s messages",
                result["message_count"],
                result["actions_done"],
            )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @imap_connection
    def move_messages_by_ids(
        self,
        message_ids: Union[str, List],
        target_folder: str,
        source_folder: str,
        use_gmail_search: bool = False,
    ) -> bool:
        """Move message by their Message-ID's from source folder to target folder

        :param message_ids: one Message-ID as string or list of Message-IDs
        :param source_folder: location of the messages, default `INBOX`
        :param target_folder: where messages should be move into
        :param use_gmail_search: set to True to use `Rfc822msgid` search, default
            is `HEADER Message-ID` search
        :return: True if all move operations succeeded, False if not
        """
        if not message_ids:
            self.logger.warning("No message ids given for Move Messages By IDs")
            return False
        idlist = message_ids if isinstance(message_ids, list) else [message_ids]
        results = []
        for mid in idlist:
            if use_gmail_search:
                move_criterion = f"gmail:Rfc822msgid:{mid}"
            else:
                move_criterion = f"HEADER Message-ID {mid}"
            result = self.move_messages(
                criterion=move_criterion,
                source_folder=source_folder,
                target_folder=target_folder,
            )
            if not result:
                self.logger.warning("Moving Message-ID '%s' failed", mid)
            results.append(result)
        return all(results)

    @imap_connection
    def _modify_gmail_labels(
        self,
        labels: str,
        criterion: Union[str, dict],
        action: bool = True,
        source_folder: str = None,
    ) -> bool:
        if isinstance(criterion, dict):
            selected_folder = source_folder or self.selected_folder
            self.select_folder(selected_folder)
            status = self._perform_actions(
                mail=criterion,
                actions=[action],
                labels=labels,
            )
            return status
        else:
            result = self._do_actions_on_messages(
                criterion=criterion,
                actions=[action],
                labels=labels,
                source_folder=source_folder,
            )

            return (
                result["actions_done"] > 0
                and result["actions_done"] == result["message_count"]
            )

    @imap_connection
    def add_gmail_labels(self, labels, criterion, source_folder: str = None) -> bool:
        """Add GMail labels to messages matching criterion and if given,
        source folder

        :param labels: comma separated list of labels to add
        :param criterion: label messages matching criterion
        :param source_folder: look for messages in this folder, default all folders
        :return: status of the operation

        Example:

        .. code-block:: robotframework

            Add Gmail Labels  customer1   SUBJECT "order confirmation"
            Add Gmail Labels  wip         SUBJECT "order confirmation"   customerfolder
        """
        return self._modify_gmail_labels(
            labels, criterion, Action.glabel_add, source_folder
        )

    @imap_connection
    def remove_gmail_labels(self, labels, criterion, source_folder: str = None) -> bool:
        """Remove GMail labels to messages matching criterion and if given,
        source folder

        :param labels: comma separated list of labels to remove
        :param criterion: unlabel messages matching criterion
        :param source_folder: look for messages in this folder, default all folders
        :return: status of the operation

        Example:

        .. code-block:: robotframework

            Remove Gmail Labels  wip  SUBJECT "order confirmation"
            Remove Gmail Labels  wip  SUBJECT "order confirmation"  customerfolder
        """
        return self._modify_gmail_labels(
            labels, criterion, Action.glabel_remove, source_folder
        )

    @imap_connection
    def do_message_actions(
        self,
        criterion: str = "",
        actions: list = None,
        source_folder: str = None,
        target_folder: str = None,
        labels: str = None,
        limit: int = None,
        overwrite: bool = False,
        prefix: str = None,
    ) -> Any:
        """Do actions to messages matching criterion and if given,
        source folder

        Actions can be:

        - msg_copy
        - msg_delete
        - msg_flag
        - msg_unflag
        - msg_read
        - msg_unread
        - msg_save
        - msg_attachment_save
        - glabel_add
        - glabel_remove

        Result object contains following attributes:

        - actions_done, number of messages on which action was performed
        - message_count, number of messages matching criterion
        - ids, message ids matching criterion
        - uids, dictionary of message uids and message content

        :param criterion: perform actions on messages matching this
        :param actions: list of actions to perform on matching messages
        :param source_folder: look for messages in this folder, default all folders
        :param target_folder: can be file path or email folder
         (for example action: msg_copy)
        :param labels: comma separated list of labels (for example action: glabel_add)
        :param limit:  maximum number of messages (for example action: msg_delete)
        :param overwrite: to control if file should overwrite
         (for example action: msg_attachment_save)
        :param prefix: prefix to be added into filename (for example: msg_save)
        :return: result object

        Example:

        .. code-block:: robotframework

            ${actions}=   Create List  msg_unflag  msg_read  msg_save  msg_attachment_save
            Do Message Actions    SUBJECT "Order confirmation"
            ...                   ${actions}
            ...                   source_folder=XXX
            ...                   target_folder=${CURDIR}
            ...                   overwrite=True
        """  # noqa: E501

        parsed_actions = [to_action(act) for act in actions]
        result = self._do_actions_on_messages(
            criterion=criterion,
            actions=parsed_actions,
            labels=labels,
            source_folder=source_folder,
            target_folder=target_folder,
            limit=limit,
            overwrite=overwrite,
            prefix=prefix,
        )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @staticmethod
    def _ensure_path_object(source: FilePath) -> Path:
        if not isinstance(source, Path):
            source = Path(source)
        return source.expanduser().resolve()

    def email_to_document(
        self, input_source: Union[FilePath, BinaryIO, bytes], output_path: FilePath
    ):
        """Convert a raw e-mail into a Word document.

        This keyword extracts the HTML (or Text) content from the passed input e-mail
        and saves it into docx format at the provided output path.

        :param input_source: Path, bytes or file-like object with the input raw e-mail
            content
        :param output_path: Where to save the output docx file

        Example:

        **Robot Framework**

        .. code-block:: robotframework

            Convert email to docx
                ${mail_file} =     Get Work Item File    mail.eml
                Email To Document    ${mail_file}    ${OUTPUT_DIR}${/}mail.docx

        **Python**

        .. code-block:: python

            from pathlib import Path
            from RPA.Email.ImapSmtp import ImapSmtp
            from RPA.Robocorp.WorkItems import WorkItems

            lib_work = WorkItems()
            lib_mail = ImapSmtp()

            def convert_email_to_docx():
                lib_work.get_input_work_item()
                mail_file = lib_work.get_work_item_file("mail.eml")
                lib_mail.email_to_document(mail_file, Path("./output") / "mail.docx")
        """

        if hasattr(input_source, "read"):
            self.logger.info("Reading raw e-mail bytes from the provided source object")
            data = input_source.read()
        elif isinstance(input_source, bytes):
            self.logger.info("Using the provided source bytes as raw e-mail content")
            data = input_source
        else:
            input_source = self._ensure_path_object(input_source)
            self.logger.info("Reading raw e-mail bytes from: %s", input_source)
            data = input_source.read_bytes()
        if not isinstance(data, bytes):
            raise ValueError(f"bytes expected for e-mail parsing, got {type(data)}")

        self.logger.info("Getting the html/text from the raw e-mail")
        message = message_from_bytes(data)
        body, _ = self.get_decoded_email_body(message, html_first=True)

        self.logger.debug("Converting html/text content:\n%s", body)
        # pylint: disable=import-outside-toplevel
        from htmldocx import HtmlToDocx

        h2d_parser = HtmlToDocx()
        docx = h2d_parser.parse_html_string(body.strip())
        output_path = self._ensure_path_object(output_path)
        self.logger.info("Writing converted document into: %s", output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        docx.save(output_path)

    def generate_oauth_string(self, username: str, access_token: str) -> str:
        """Generate and return an OAuth2 string compatible with the IMAP/POP/SMTP
        XOAUTH2 protocol.

        This string usually gets passed to the ``Authorize`` keyword as `password` when
        `is_oauth=${True}`.

        :param username: The e-mail address you're going to send the e-mail with.
        :param access_token: Access token string found in the dictionary obtained with
             ``Get OAuth Token`` or ``Refresh OAuth Token``.
        :returns: Base64 encoded string packing these credentials and replacing the
            legacy `password` when enabling the OAuth2 flow.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Authorize ImapSmtp
                ${password} =   Generate OAuth String    ${username}
                ...    ${token}[access_token]
                Authorize    account=${username}    is_oauth=${True}
                ...     password=${password}
        """
        auth_string = f"user={username}\1auth=Bearer {access_token}\1\1"
        return base64.b64encode(auth_string.encode()).decode()

    def convert_eml_file_into_message(
        self, eml_filepath: str, save_attachments_directory: str = None
    ):
        """Converts EML file into message dictionary.

        Returned dictionary contains:

        - **headers** of the email
        - **attachments** the filenames of the attachments or if attachments
          have been saved then they are absolute filepaths to each attachment
        - **body_text** is the TEXT formatted content of the email body
        - **body_html** is the HTML formatted content of the email body

        :param eml_filepath: filepath to the EML file
        :param save_attachments_directory: path to the directory where possible
         attachments will be saved to, if not given then attachment filenames are
         returned in a list of the return dictionary in the key 'attachments'
        :return: dictionary containing information aboutthe EML message
        """
        message = None
        message_dict = {
            "headers": None,
            "attachments": None,
            "body_text": None,
            "body_html": None,
        }

        with open(eml_filepath, "rb") as f:
            message = message_from_binary_file(f)
        if message:
            # Get the headers
            headers = {}
            for key, val in message.items():
                headers[key] = val
            message_dict["headers"] = headers

            # Get the body
            get_email_body_from_message(message, message_dict)
            # Get the attachments
            get_attachments_from_message(
                message, save_attachments_directory, message_dict
            )
        return message_dict
